#include <stdio.h>
#include <string.h>
#include "mfhdf.h"
#include "cdf.h"
#include "hfile.h"

/******************************************************************************
   This program a translates a HDF4 file into a CDF file, and it supports
   the following HDF4 objects:

     - AN (Annotation)
       * File labels (AN_FILE_LABEL) => converted to CDF global attributes
       * File descriptions (AN_FILE_DESC) => converted to CDF global attributes

     - Vgroup (including nested Vgroups)
       * Vgroup name => converted to a CDF global attribute
       * Vgroup attributes => converted to the global attribute entries
       * Data labels (AN_DATA_LABEL) 
             => converted to CDF global attribute entries
       * Data descriptions (AN_DATA_DESC) 
             => converted to CDF global attribute entries 

     - Vdata 
       * Vdata name => converted to a CDF global attribute
       * Vdata attributes => converted to the global attribute entries
       * Data labels (AN_DATA_LABEL)
             => converted to CDF global attribute entries
       * Data descriptions (AN_DATA_DESC)
             => converted to CDF global attribute entries
       * Each field in Vdata => converted to a CDF variable
       * Each field's attribute => converted to CDF a variable attribute

     - Scientific Data Set (SDS)
       * SDS file attribute => converted to CDF a global attribute
       * Each dataset in a SDS => converted to a CDF variable
       * SDS attribute => converted to CDF a variable attribute

     - General Raster (GR)
       * GR file attribute => converted to CDF a global attribute
       * Raster iamge (8, 24, or general) attribute
             => converted to a CDF variable attribute
       * 8-bit raster image => converted to a CDF variable
       * 24-bit raster image => converted to a CDF variable 
       * General raster image => converted to a CDF variable 
       * Palette => converted to a CDF variable attribute

   NOTE:
      This program also allows users to rename HDF4 attribute and variable
      manes by defining the mapping information in the mapping file
      (hdf4_to_cdf_mapping.dat). 

******************************************************************************/

#define FALSE             0
#define MAX_SDS           6000    /* Max # of SDSs in a HDF4 file          */
#define MAX_VGROUPS       6000    /* Max # of Vgroups in a HDF4 file       */
#define MAX_RASTER_IMAGES 6000    /* Max # of raster images in a HDF4 file */

#define MAX_ATTRS         10000   /* Max # of HDF4 attributes in a HDF4 file */
#define MAX_VARS          13000   /* Max # of HDF4 variables  in a HDF4 file */
#define MAX_MAPPING       900     /* Max # of mappings                       */

#define DEFAULT_MAPPING_FILE   "hdf4_to_cdf_mapping.dat"
#define MAX_REC_LENGTH     180  /* Max record length of DEFAULT_MAPPING_FILE */
#define MAX_FILE_NAME_LEN  150    /* Max. file name length                   */
#define UNKNOWN_DATATYPE   -99    /* Unsupported datatype indicator        */

/**********************/
/*  Global variables  */
/**********************/
CDFid    id;                   /* pointer to the CDF output file */

char           *charPtr;
short          *shortPtr;
long           *longPtr;
float          *floatPtr;
double         *doublePtr;
unsigned char  *ucharPtr;
unsigned short *ushortPtr;
unsigned long  *ulongPtr;

char     *filename;            /* HDF4 input file name */
int32    n_Vgroup_IDs;         /* # of Vgroup IDs in the input file          */
int32    n_SDSs_processed;     /* Total # of scientific data sets processed  */
int32    n_rasters_processed;  /* Total # of raster images processed (8-bit, */
                               /* 24-bit, and general raster)                */

int32    vgroup_ids[MAX_VGROUPS];
int32    processed_sds_indicies[MAX_SDS];
uint16   processed_rasters_indicies[MAX_RASTER_IMAGES]; 
             /* Processed raster image indicies (8-bit, 24-bit, and general) */ 
char *nomsg = "No next argument";

/*************************************************************************/
/*   Arrays to hold the HDF4-to-CDF mapping information that are loaded  */
/*   from the default or user-define mapping file.                       */
/*************************************************************************/
char hdf4AttrNames[MAX_MAPPING][MAX_NC_NAME];
char cdfAttrNames[MAX_MAPPING][CDF_ATTR_NAME_LEN256+1];
char comments[MAX_MAPPING][81];

char  output_filename[MAX_FILE_NAME_LEN];    /* HDF4-to-CDF output file name */ 

char  mappingFile[MAX_FILE_NAME_LEN];  /* HDF4-to-CDF Mapping file name */
int   n_mapping_attrs;           /* # of attributes in the mapping file */

struct cdfvar {           /* CDF variable structure */
     char name[CDF_VAR_NAME_LEN256+1];   /* variable name                */
     long num;                        /* variable number              */
     long datatype;
     long dimensionality;             /* variable dimensionality      */
     long dimSizes[CDF_MAX_DIMS];     /* variable dimension sizes     */
     long numValues;                  /* # of data values in dimSizes */
     long recVariance;                /* record variance              */
};

struct cdfattr {         /* CDF attribute structure */
    char name[CDF_ATTR_NAME_LEN256+1];   /* attribute name   */
    long num;                         /* attribute number */
    long entryNum;                    /* attribute number */
    long datatype;                    /* attribute datatype */
    long scope;                       /* attribute scope  */
}; 


struct cdfattr_struct {          /* CDF attribute structure */
    char name[CDF_ATTR_NAME_LEN256+1];   /* attribute name  */
    long scope;                       /* attribute scope */
    int  seq_no;                      /* attribute sequence number */
} processed_cdf_attrs[MAX_ATTRS];
int  n_cdf_attrs_created = 0;

struct cdfvar_struct {      /* CDF attribute structure */
    char name[CDF_VAR_NAME_LEN256+1];   /* variable name            */
    int  seq_no;                     /* variable sequence number */
} processed_cdf_vars[MAX_VARS];
int  n_cdf_vars_created = 0;

struct VD_field {                    /* structure for Vdata field */
     char   name[CDF_VAR_NAME_LEN256];
     int32  datatype;
     int32  numValues;       /* # of values in the field         */
};

struct VD_attr {         /* structure for Vdata or Vdata field attribute */
     char   name[CDF_VAR_NAME_LEN256];
     int32  datatype;
     int32  n_values;        /* # of values in this attribute    */
     int32  size;            /* Size, in bytes, of the attribute */
};

int DEBUG = FALSE;
int DEBUG_DATA = FALSE;
int COMPRESS = TRUE;

/**************************/
/*  Function definitions  */
/**************************/
void h4error (char *);
char * hdf4_datatype (int32);
long get_cdf_datatype (int32);
void print_SDgetinfo(int32, char *, int32, int32 *, int32, int32);
void print_GRgetiminfo (int32, char *, int32,int32,int32,int32 *,int32);
void print_raster_image_data (char *, long, uint8 *);
void get_sds_info (int32, int32, int32 *, int32 *, int32 *);
void allocate_memory (int32, int32, int32);
void read_sds_data (int32,char *,int32,int32,int32 *,int32,int32, struct cdfvar);
void read_sds_attr_data (int32, int32, int32, int32, int32, struct cdfattr, char *);
void read_raster_data (int32, int32,int32,int32 *,int32, int32, struct cdfvar);
void read_palette_data (int32, int32, int32, int32, char *, struct cdfattr *); 
void read_gr_attr_data (int32, int32, int32, int32, int32, struct cdfattr, char *);
void read_vdata_data (int32, int32, int32, int32, char *, struct cdfvar);
void print_hdf4_data (int32, int32, int32);
void read_AN_file_annotation (int32, int32, ann_type);
void read_data_annotation (int32, ann_type, uint16, uint16, struct cdfattr *);
void read_annotation (int32, int32, struct cdfattr *);
char * str_annotation_type (ann_type);
void process_Vgroup (int32, int32, int32, int32, int32, char *);
void process_SDS (int32, int32);
void process_GR (int32, int32);
void process_palette (int32, char *, struct cdfattr *);
void process_VD (int32, int32, int32, char *);
void get_Vgroup_IDs (int32);
void process_vdata_attr (int32, int32, int32, char *, struct cdfattr *);
void read_vdata_attr_data (int32, int32, intn, struct VD_attr, char *, struct cdfattr *);
void process_vgroup_attr (int32, struct cdfattr *);
void read_vgroup_attr_data (int32, intn, int32, int32, struct cdfattr *);
void process_SDS_attrs (int32, int32, long, char *);
void process_GR_attrs (int32, int32, long, char *);
void process_AN_DATA_objects (int32, uint16, int32, struct cdfattr *);
void usage();
void showVersions();
void parseCmdLine (int, char *[]);
void errorMsg (char *);
char * getNextArgument (int, char *[], int);
void load_mapping_info();
void map_HDF4_name_to_CDF (char *, char *);
void removeFilepath();
char * removeFileExtension(char *, char *);
int endsWith(char *, char *);
char * strlaststr(char *, char *);

void cdf_create_file(char *);
long cdf_create_var (struct cdfvar *, int32, int32, int32 *, long);
long cdf_create_attribute (char *, long);
void cdf_create_placeholder_attributes();
void cdf_get_new_variable_name (char *, int *);
void cdf_get_new_attribute_name (char *, long, int *, int *);
void cdf_create_gAttribute_entry (long, long, long, long, void *);
void cdf_create_vAttribute_entry (char *, char *, long, long, void *);
void cdf_status_handler (CDFstatus, char *);
void cdf_write_vardata (long, long, long, long, long [], void *);
void cdf_add_gAttribute_entry (struct cdfattr *, char *, char *);
void cdf_allocate_zvar_records (long, long);
void cdf_handle_attrs_in_mapping_file (char *);
char *  cdf_str_datatype (long);
int  cdf_setCompression (struct cdfvar, int);

/**********/
/*  Main  */
/**********/
int main (int argc, char *argv[])
{
   int32    file_id, sd_id, sds_index, raster_index, an_id, gr_id;
   int32    n_file_labels, n_file_descs,  n_data_labels, n_data_descs;
   int32    n_sds_datasets, n_sds_file_attrs, n_gr_file_attrs;
   int32    n_raster_images;
   int32    n_lone_vgs, lone_vg_number;
   int32    n_lone_vds, lone_vd_number;
   int32    *ref_array;       /* array to hold reference numbers */
   intn     status;           /* SUCCEED or FAIL */
   intn     n_raster8_images, n_raster24_images;
   int      i, sds_already_processed, raster_already_processed; 

   /**********************/
   /*  CDF declaration   */
   /**********************/
   CDFstatus   cstatus;           /* CDF status code */
   char        cdf_filename[160], temp_filename[160];      

   /***************************************************************/
   /* Process the command line input  and the mapping information */
   /***************************************************************/
   if (argc <= 1)
       usage();             /* HDF4 input file name not specified */
   else {
        filename = argv[argc-1];                 /* Get the input file name */
        strcpy(mappingFile, "UNSPECIFIED");
        strcpy(output_filename, "UNSPECIFIED");
        parseCmdLine(argc, argv);
        if (strcmp(mappingFile, filename) == 0)
            errorMsg("** Error - either input file or mapping file is missing");
        if (strcmp(mappingFile,"UNSPECIFIED") == 0)
            strcpy(mappingFile, DEFAULT_MAPPING_FILE);
   }

   load_mapping_info();     /* Load the HDF4-to-CDF mapping information */

   /* Remove the file path if it exists (/home/mydata.hdf => mydata.hdf). */
    printf ("\nInput file name: %s\n", filename);

    /* Strip off the hdf4 file extension before creating a CDF file. */
    strcpy (cdf_filename, filename);

    if (strcmp(output_filename, "UNSPECIFIED") == 0)
        removeFilepath(cdf_filename);
    else
        strcpy (cdf_filename, output_filename);

    if (endsWith(cdf_filename,".hdf")) 
        removeFileExtension(cdf_filename,".hdf");
    else if (endsWith(cdf_filename,".h4")) 
        removeFileExtension(cdf_filename, ".h4");
    else if (endsWith(cdf_filename,".cdf")) 
        removeFileExtension(cdf_filename,".cdf");

    cdf_create_file (cdf_filename);
    cdf_create_placeholder_attributes();

    printf ("Output file name: %s.cdf\n", cdf_filename);

   /*********************************/
   /*  Initialize global variables  */
   /*********************************/
   n_SDSs_processed = 0;        /* Total # of SDSs processed               */
   n_rasters_processed = 0;     /* Total # of all raster images processed  */
   n_Vgroup_IDs = 0;            /* # of Vgroup IDs in the file             */

   /*****************************************************/
   /* Process the given HDF4 file and create a CDF file */
   /*****************************************************/
   file_id = Hopen (filename, DFACC_READ, 0);
   if (file_id == FAIL) {
       h4error ("Hopen");
       exit (1);
   }
   status  = Vstart (file_id);                /* Initialize Vdata and Vgroup */
   an_id   = ANstart (file_id);               /* Initialize Annotation       */
   gr_id   = GRstart (file_id);               /* General raster images       */
   sd_id   = SDstart(filename, DFACC_READ);   /* Initialize SDS              */

   /***********************************************************/
   /* Get the number of file and data labels and descriptions */ 
   /***********************************************************/
   status = ANfileinfo (an_id, &n_file_labels, &n_file_descs, 
                        &n_data_labels, &n_data_descs); 
   if (status == FAIL) h4error("ANfileinfo");
   if (DEBUG) 
       printf ("\n*** # of AN_FILE_LABEL= %d, AN_FILE_DESC= %d, "
               "data labels= %d, data descs=%d\n\n", 
               n_file_labels, n_file_descs, n_data_labels, n_data_descs);

   read_AN_file_annotation (an_id, n_file_labels, AN_FILE_LABEL);
   read_AN_file_annotation (an_id, n_file_descs, AN_FILE_DESC); 
    
   /*************************************************************************/
   /* Get the number of SDS datasets and process the SDS global file        */
   /* attributes.                                                           */
   /*                                                                       */
   /* NOTE:                                                                 */
   /*    The global file attributes written by the SD interface             */
   /*    are only visible to the SD interface, and they are not visible to  */
   /*    the GR global file attribute routine (albeit they are global file  */
   /*    attributes.                                                        */ 
   /*************************************************************************/
   status = SDfileinfo (sd_id, &n_sds_datasets, &n_sds_file_attrs);
   if (status == FAIL) h4error("SDfileinfo");
   if (DEBUG) 
       printf ("\n*** # of SDS datasets = %d,  SDS file attrs = %d\n", 
               n_sds_datasets, n_sds_file_attrs);

   /* Process the SDS HDF4 file attributes */
   if (n_sds_file_attrs > 0)
       process_SDS_attrs (sd_id, n_sds_file_attrs, 
                          GLOBAL_SCOPE,    /* CDF attribute scope */
                          "");             /* CDF variable name   */

   /*************************************************************************/
   /* Get the number of raster images and process the GR global file        */
   /* attributes.                                                           */
   /*                                                                       */
   /* NOTE:                                                                 */
   /*    The global file attributes written by the GR interface             */
   /*    are only visible to the GR interface, and they are not visible to  */
   /*    the SD global file attribute routine (albeit they are global file  */
   /*    attributes.                                                        */ 
   /*************************************************************************/
   status = GRfileinfo (gr_id, &n_raster_images, &n_gr_file_attrs);
   if (status == FAIL) h4error("GRfileinfo");
   if (DEBUG)
       printf ("\n*** # of raster images = %d,  global file attrs = %d\n",
               n_raster_images, n_gr_file_attrs);

   n_raster8_images = DFR8nimages (filename);      
   if (n_raster8_images == FAIL) h4error ("DFR8nimages");

   n_raster24_images = DF24nimages (filename);
   if (n_raster24_images == FAIL) h4error ("DF24nimages");

   if (DEBUG) {
       printf ("\n    #  8-bit raster images = %d\n", n_raster8_images);
       printf ("    # of 24-bit raster images = %d\n", n_raster24_images);
   }
   if (n_gr_file_attrs > 0)        /* Process the GR HDF4 file attributes */ 
       process_GR_attrs (gr_id, n_gr_file_attrs, GLOBAL_SCOPE,
                         "");      /* dummy raster image name */

   /****************************/
   /* Process the lone Vgroups */
   /****************************/
   n_lone_vgs = 0;         /* If not set, an error will occur */
   n_lone_vgs = Vlone (file_id, NULL, n_lone_vgs);
   if (DEBUG) printf ("\n**** # of lone Vgroups = %d\n", n_lone_vgs);

   if (n_lone_vgs > 0) {
       ref_array = (int32 *) malloc (sizeof(int32) * n_lone_vgs);
       /* Get the reference numbers for all the lone vgroups */
       n_lone_vgs = Vlone (file_id, ref_array, n_lone_vgs);
       if (n_lone_vgs == FAIL) h4error ("Vlone");

       for (lone_vg_number=0; lone_vg_number < n_lone_vgs; lone_vg_number++) {
            if (DEBUG) 
                printf ("\n  In main, processing the lone Vgroup #%d...\n",
                        lone_vg_number);

            process_Vgroup (file_id, sd_id, an_id, gr_id,
                            ref_array[lone_vg_number], 
                            filename);
       }
       free (ref_array);
   }

   /***************************/
   /* Process the lone Vdatas */
   /***************************/
   n_lone_vds = 0;         /* If not set, an error will occur */
   n_lone_vds = VSlone (file_id, NULL, n_lone_vds);
   if (DEBUG) printf ("\n**** # of lone Vdatas = %d\n", n_lone_vds);

   if (n_lone_vds > 0) {
       /* Get the reference numbers for all the lone vdatas */
       ref_array = (int32 *) malloc (sizeof(int32) * n_lone_vds);
       n_lone_vds = VSlone (file_id, ref_array, n_lone_vds);
       if (n_lone_vds == FAIL) h4error ("VSlone");

       for (lone_vd_number=0; lone_vd_number < n_lone_vds; lone_vd_number++) {
            process_VD (file_id, an_id, ref_array[lone_vd_number], "");
       }
       free (ref_array);
   }
   if (DEBUG) printf ("\n");

   /* get_Vgroup_IDs (file_id);     */
   /* process_VDs (file_id, an_id); */

   /************************************************************/
   /* Process the lone SDSs - SDSs don't belong to any Vgroups */
   /************************************************************/
   if (DEBUG) printf ("\n---- Lone SDSs\n");

   for (sds_index=0; sds_index < n_sds_datasets; sds_index++) {
        sds_already_processed = FALSE;
        for (i=0; i < n_SDSs_processed; i++) {
             if (sds_index == processed_sds_indicies[i]) {
                 sds_already_processed = TRUE;
                 break;
             }
        }
        if (!sds_already_processed) process_SDS (sd_id, sds_index);
   }

   /**********************************************************************/
   /* Process the lone general raster images that are 8-bit, 24-bit, and */
   /* general raster images.                                             */
   /**********************************************************************/
   if (DEBUG) printf ("\n---- Lone raster images\n");

   for (raster_index=0; raster_index < n_raster_images; raster_index++) {
        /************************************************************/
        /* Determine whether this image has been already processed. */
        /************************************************************/
        raster_already_processed = FALSE;
        for (i=0; i < n_rasters_processed; i++) {
             if (raster_index == processed_rasters_indicies[i]) {
                 raster_already_processed = TRUE;
                 break;
             }
        }
        if (!raster_already_processed) process_GR (gr_id, raster_index);
        if (DEBUG) printf ("\n");
   }

   /*******************/
   /* Clean up (HDF4) */
   /*******************/
   status = SDend (sd_id); 
   status = Vend (file_id);
   status = ANend (an_id);
   status = GRend (gr_id);
   status = Hclose (file_id);   

    /* Close the CDF output file */
    cstatus = CDFlib (CLOSE_, CDF_, NULL_);
    if (cstatus != CDF_OK) cdf_status_handler (cstatus, "CLOSE_, CDF_");
    exit (0);
}


/*************************************************************************/
/*  Process all the HDF4 objects in the given Vgroup (referenced by the  */
/*  reference number).  If an object is Vgroup within a Vgroup, then a   */
/*  recursion is performed to handle nested Vgroups.                     */
/*************************************************************************/
void process_Vgroup (int32 file_id, int32 sd_id, int32 an_id, int32 gr_id,
                     int32 vgroup_ref_num, 
                     char *filename)
{
   int32    index, vgroup_id, vdata_id, sds_index, raster_image_index;
   int32    n_tagrefs, obj_tag, obj_ref;
   int      create_new_attr, new_attr_name_created;
   intn     status;           /* SUCCEED or FAIL */
   uint16   vgroup_tag;
   char     vgroup_name[VGNAMELENMAX], vgroup_class_name[VGNAMELENMAX];
   struct cdfattr  cdfAttr;

   vgroup_id = Vattach (file_id, vgroup_ref_num, "r");
   status = Vgetname (vgroup_id, vgroup_name);
   status = Vgetclass (vgroup_id, vgroup_class_name);

   if (DEBUG) {
       printf ("\n**** VGroup name= %s, class name= %s, filename= %s\n",
               vgroup_name, vgroup_class_name, filename);
       fflush (stdout);
   }

   /**********************************************************************/
   /*  If the name of the vgroup is the input HDF4 file name, skip this  */
   /*  group since it contains information that are only relevant to the */
   /*  HDF4 library.                                                     */ 
   /**********************************************************************/
   /* if (strcmp(vgroup_name, filename) == 0) return;  */


   /***********************************************************************/
   /* If the vgroup's class name is "RI0.0", this group only contains the */
   /* GR file global attributes.  SKip this group since these global      */
   /* file attributes are already processed in main.                      */ 
   /***********************************************************************/
   if (strcmp(vgroup_class_name,"RIG0.0") == 0) return;  

   /*************************************************************************/ 
   /* Create a global attribute for this Vgroup.  All the Vgroup attributes */
   /* will be captured as entries of the global atrributes.                 */
   /*************************************************************************/ 
   strcpy (cdfAttr.name, "Vgroup_");
   strcat (cdfAttr.name, vgroup_name);
   cdf_get_new_attribute_name (cdfAttr.name, GLOBAL_SCOPE,
                               &create_new_attr, &new_attr_name_created);
   if (create_new_attr) {
       cdfAttr.num = cdf_create_attribute (cdfAttr.name, GLOBAL_SCOPE);
       cdfAttr.entryNum = -1;
       cdf_add_gAttribute_entry (&cdfAttr, "Original Vgroup name = ",
                                 vgroup_name);
       cdfAttr.scope = GLOBAL_SCOPE;
   }

   /***********************************************************************/
   /*  Get the number of HDF4 objects in this Vgroup.  An HDF4 object is  */
   /*  represented by a tag and reference number.                         */
   /*******************************************.***************************/
   n_tagrefs = Vntagrefs (vgroup_id);

   if (DEBUG) {
       printf ("  --- # of objects in this group = %d\n", n_tagrefs);
       fflush (stdout);
   }

   /********************************************************************/
   /*  Process the annotation objects (AN_DATA_LABEL and AN_DATA_DESC) */
   /*  in this Vgroup if there are any.                                */
   /********************************************************************/
   vgroup_tag = DFTAG_VG;
   process_AN_DATA_objects (an_id, vgroup_tag, vgroup_ref_num, &cdfAttr);

   /**************************************************************/ 
   /*  Process the attributes that are attached to this Vgroup.  */
   /**************************************************************/ 
   process_vgroup_attr (vgroup_id, &cdfAttr);

   /******************************************/
   /* Process all the objects in this Vgroup */
   /******************************************/
   for (index=0; index < n_tagrefs; index++) {
        status = Vgettagref (vgroup_id, 
                             index,             /* Object location in Vgroup */
                             &obj_tag, &obj_ref);
        if (status == FAIL) h4error ("Vgettagref");

        if (DEBUG) {
            printf ("\n  Checking for the object #%d in %s...\n", 
                    index, vgroup_name);
            fflush (stdout);
        }
        if (Visvg(vgroup_id, obj_ref)) {        /* Vgroup */
            
            if (DEBUG) printf ("\n%d) obj_ref %d = Vgroup\n", index, obj_ref);

            /* Recursion */
            if (DEBUG) {
                printf ("   !!!! Calling recursion...\n");
                fflush (stdout);
            }
            process_Vgroup (file_id, sd_id, an_id, gr_id, obj_ref, filename); 

            if (DEBUG) {
                printf ("   !!!! After recursion...\n");
                fflush (stdout);
            }
        }

        else if (Visvs(vgroup_id, obj_ref)) {         /* Vdata */
            if (DEBUG) {
                printf ("\n%d) obj_ref %d = Vdata\n", index, obj_ref);
                fflush (stdout);
            }
            process_VD (file_id, an_id, obj_ref, vgroup_name); 
        }

        else if (obj_tag == DFTAG_NDG) {              /* Scientific Data Set */
            if (DEBUG) {
                printf ("\n%d) obj_tag %d = SDS\n", index, obj_tag);
                fflush (stdout);
            }
            sds_index = SDreftoindex(sd_id, obj_ref);
            processed_sds_indicies[n_SDSs_processed++] = sds_index;
            process_SDS (sd_id, sds_index);
        }

        /*****************************************************************/
        /* Process a 8-bit, 24-bit, or a general raster image.  The      */
        /* DFTAG_RIG tag is a generic tag to represent a raster image.   */
        /* Actual image type is determined in Process_GR.                */
        /*                                                               */
        /* NOTE:                                                         */
        /*    DFTAG_RI8 is an obsolete tag used for representing the     */
        /*    8-bit raster image.  Handle this tag if it's used.         */
        /*****************************************************************/
        else if (obj_tag == DFTAG_RIG || obj_tag == DFTAG_RI8) {
             if (DEBUG) {
                 printf ("\n%d) obj_ref %d = Raster\n",index,obj_ref);
                 fflush (stdout);
             }
             raster_image_index = GRreftoindex (gr_id, (uint16) obj_ref);
             if (raster_image_index == FAIL) h4error ("GRreftoindex"); 
             process_GR (gr_id, raster_image_index);
        }
   }
   status = Vdetach (vgroup_id); 
   if (DEBUG) {
       printf ("end of processVgroup\n");
       fflush (stdout);
   }
}


/*************************************************************************/
/*  Process the given Scientific Data Set (SDS) referenced by the data   */
/*  set index.  The index is a location within a SDS.  Note that there   */ 
/*  be one or more data sets in a SDS.                                   */ 
/*************************************************************************/
void process_SDS (int32 sd_id, int32 sds_index)
{
    int32    sds_id, n_sds_attrs, hdf4_type;
    int32    rank, dim_sizes[CDF_MAX_DIMS];
    int32    numRecs, numElements;
    char     sds_name[MAX_NC_NAME];
    long     record_variable;
    intn     status;           /* SUCCEED or FAIL */
    int      var_already_exists;
    struct cdfvar   cdfVar;


    /* Get the SDS dataset information */
    sds_id = SDselect(sd_id, sds_index);
    status = SDgetinfo(sds_id, sds_name, &rank, dim_sizes, &hdf4_type,
                       &n_sds_attrs);
    if (DEBUG) {
        print_SDgetinfo(sds_id, sds_name, rank, dim_sizes, hdf4_type, 
                        n_sds_attrs);
       fflush (stdout);
    }

    /* Get the number of values on each record and records to read */
    get_sds_info (sds_id, rank, dim_sizes, &numElements, &numRecs);
    if (DEBUG) 
        printf ("    numRecs= %d, arraySize= %d\n", numRecs, numElements);

    allocate_memory (hdf4_type, numElements, numRecs);

    if (SDisrecord(sds_id)) {        /* record varying */
        record_variable = TRUE;
        cdfVar.recVariance = TRUE;
    }
    else {
        record_variable = FALSE;
        cdfVar.recVariance = FALSE;
    }

    /*************************************************/
    /* Ensure that there are no duplicate variables  */
    /*************************************************/
    strcpy (cdfVar.name, sds_name);
    cdf_get_new_variable_name (cdfVar.name, &var_already_exists);

    cdfVar.num = cdf_create_var (&cdfVar, hdf4_type, rank, dim_sizes, 
                                 record_variable);
    if (cdfVar.num == UNKNOWN_DATATYPE) {
        printf ("Unknown HDF4 data type (process_SDS): %d\n", hdf4_type);
        return;         /* skip the current data set */
    }

    if (COMPRESS) cdf_setCompression (cdfVar, numRecs);
    read_sds_data (sds_id, sds_name, hdf4_type, rank, dim_sizes, 
                   numElements, numRecs, cdfVar);

    /*  Process the SDS attributes */
    if (n_sds_attrs > 0)
        process_SDS_attrs (sds_id, n_sds_attrs, 
                           VARIABLE_SCOPE,     /* CDF attribute scope */
                           cdfVar.name);       /* CDF variable name   */

    /**************************************************************/
    /* Create a variable attribute and its entry if one is        */
    /* defined in the mapping file.                               */
    /**************************************************************/
    cdf_handle_attrs_in_mapping_file (cdfVar.name);

    status = SDendaccess(sds_id);
}


/**************************************************************************/
/*  Process the given raster image referenced by the raster image index.  */
/*  The index is a location within a General Raster (GR) set.  Note that  */
/*  there can be one or more raster images in a Raster Image Group (RIG). */ 
/**************************************************************************/
void process_GR (int32 gr_id, int32 raster_index)
{
    int32    raster_id, n_raster_attrs, n_pixel_components, hdf4_type, rank;
    int32    interlace_mode, i; 
    int32    dim_sizes[CDF_MAX_DIMS], temp_dim_sizes[CDF_MAX_DIMS];
    int32    palette_id, palette_index;
    int32    numRecs, numElements;
    uint16   obj_ref;
    char     raster_name[H4_MAX_GR_NAME];
    long     record_variable;
    intn     status, n_palettes;           /* SUCCEED or FAIL */
    int      var_already_exists, create_new_attr, new_attr_name_created;
    struct cdfvar   cdfVar;
    struct cdfattr  cdfAttr;

    /************************************/
    /* Get the raster image information */
    /************************************/
    raster_id = GRselect(gr_id, raster_index);

    status = GRgetiminfo (raster_id, raster_name, &n_pixel_components, 
                          &hdf4_type, &interlace_mode, dim_sizes, 
                          &n_raster_attrs);
    if (status == FAIL) h4error ("GRgetiminfo");

    /* Get the image reference number */
    obj_ref = GRidtoref(raster_id);

    if (DEBUG) {
        printf ("obj ref = %d\n", obj_ref);
        print_GRgetiminfo (raster_id, raster_name, n_pixel_components, 
                           hdf4_type, interlace_mode, dim_sizes, 
                           n_raster_attrs);
        fflush (stdout);
    }

    /*************************************************/
    /* Ensure that there are no duplicate variables  */
    /*************************************************/
    strcpy (cdfVar.name, raster_name);
    cdf_get_new_variable_name (cdfVar.name, &var_already_exists);

    /* If the image is not a 8-bit raster image, the underlying data is     */
    /* sored as 3-D ([row, col, pixel_depth]) although the HDF4 library     */
    /* treats the image dimension as a 2-D array ([row, col]).  Before      */
    /* creating CDF variable, capture the pixel depth.                      */
    rank = 2;
    for (i=0; i < rank; i++)
         temp_dim_sizes[i] = dim_sizes[i];

    if (n_pixel_components > 1)    /* greater than 8-bit image */ 
        temp_dim_sizes[rank-1] *= n_pixel_components;

    record_variable = FALSE;       /* Each image has only one record */
    cdfVar.recVariance = FALSE;    /* every raster image has only 1 record */
    cdfVar.num = cdf_create_var (&cdfVar, hdf4_type, rank, temp_dim_sizes,
                                 record_variable);
    if (cdfVar.num == UNKNOWN_DATATYPE) {
        printf ("Unknown HDF4 data type (process_GR): %d\n", hdf4_type);
        return;         /* skip the current data set */
    }

    /**********************************************************************/
    /* Keep the processed raster image reference number so that it's not  */
    /* processed again in 'main' when it goes to process the lone images. */
    /**********************************************************************/
    processed_rasters_indicies[n_rasters_processed++] = raster_index;

    /*****************************************************/
    /*  Get the # of palettes associated with this image */
    /*****************************************************/
    n_palettes = GRgetnluts (raster_id);
    if (n_palettes == FAIL) h4error ("GRgetnluts");
    if (DEBUG) printf ("  # of palettes = %d\n", n_palettes);

    /****************************************/
    /*  Process the raster image attributes */
    /****************************************/
    if (n_raster_attrs > 0)
        process_GR_attrs (raster_id, n_raster_attrs, VARIABLE_SCOPE, 
                          cdfVar.name);

    /**********************************/
    /*  Process the Raster image data */
    /**********************************/
    numRecs = 1;
    numElements = dim_sizes[0] * dim_sizes[1] * n_pixel_components;
    if (COMPRESS) cdf_setCompression (cdfVar, numRecs);

    allocate_memory (hdf4_type, numElements, numRecs);
    read_raster_data (raster_id, hdf4_type, interlace_mode, dim_sizes, 
                      numElements, numRecs, cdfVar);

    /********************************************************************/
    /*  Process the palette data that is/are associated with this image */
    /********************************************************************/
    if (n_palettes > 0) {
        strcpy (cdfAttr.name, "Raster_image_palette");
        cdf_get_new_attribute_name (cdfAttr.name, VARIABLE_SCOPE, 
                                    &create_new_attr, &new_attr_name_created);
        if (create_new_attr)
            cdfAttr.num = cdf_create_attribute (cdfAttr.name, VARIABLE_SCOPE);
    }

    for (palette_index=0; palette_index < n_palettes; palette_index++) {
         palette_id = GRgetlutid (raster_id, palette_index);
         process_palette (palette_id, cdfVar.name, &cdfAttr);
    }

    status = GRendaccess(raster_id);
}


/**************************************************************************/
/*  Process the palette (a.k.a color lookup table) that's associate with  */
/*  the given palette ID.                                                 */
/**************************************************************************/
void process_palette (int32 palette_id,
                      char  *cdf_var_name,
                      struct cdfattr *cdfAttr)
{
    int32    n_pixel_components, hdf4_type, n_entries;
    int32    interlace_mode;
    int32    numRecs, numElements;
    intn     status;           /* SUCCEED or FAIL */

    status = GRgetlutinfo (palette_id, &n_pixel_components,
                           &hdf4_type, &interlace_mode, 
                           &n_entries);       /* # of colors in the palette */
    if (status == FAIL) h4error ("GRgetiminfo");

    if (DEBUG) 
        printf ("\nPalette: %d pixel components; %d entries\n", 
                n_pixel_components, n_entries);

    
    /*  Process the Raster image data */
    cdfAttr->datatype = get_cdf_datatype (hdf4_type);
    if (cdfAttr->datatype == UNKNOWN_DATATYPE) {
        printf ("Unknown HDF4 palette data type (process_palette): %d\n", 
                hdf4_type);
        return;         /* skip the current palette */
    }

    numRecs = 1;
    numElements = n_entries * n_pixel_components;
    allocate_memory (hdf4_type, numElements, numRecs);

    /* Set the interlace mode prior reading the palette data */
    status = GRreqlutil (palette_id, (intn) interlace_mode);
 
    read_palette_data (palette_id, hdf4_type, n_entries, numRecs,
                       cdf_var_name, cdfAttr);
}


/*************************************************************************/
/*  Process the given Vdata referenced by the Vdata reference number.    */
/*************************************************************************/
void process_VD (int32 file_id, int32 an_id, int32 vdata_ref, char *vgroup_name)
{
    int32  vdata_id, n_records, n_fields, interlace_mode, vdata_size;
    int32  n_vdata_attrs, n_vdata_field_attrs;
    int32  hdf4_dimensionality, hdf4_dim_sizes[2];
    uint16 vdata_tag;
    intn   status, n_data_labels, n_data_descs, attr_index;   
    int32  index;
    int    var_already_exists, lone_vdata, create_new_attr, 
           new_attr_name_created;
    long   record_variable, vdata_attr_num, vgroup_attr_num, attr_num;
    char   vdata_name[VSNAMELENMAX], vdata_class_name[VSNAMELENMAX];
    char   fieldname_list[2500];
    char   format_attr_name[CDF_ATTR_NAME_LEN256+1];
    char   vgroup_attr_name[CDF_ATTR_NAME_LEN256+1];
    char   vdata_attr_name[CDF_ATTR_NAME_LEN256+1];
    struct VD_field  field;
    struct cdfvar    cdfVar;
    struct cdfattr   cdfAttr;

    vdata_id = VSattach (file_id, vdata_ref, "r");

    /********************************************************************/
    /*  NOTE:                                                           */
    /*      In HDF4, vdata attributes and vdata field attributes are    */ 
    /*      stored as vdatas.  The VSisattr routine determines whether  */
    /*      a particular data is vdata attribute or not.                */ 
    /********************************************************************/
    if (VSisattr(vdata_id) != TRUE) {       /* Vdata */
        /*****************************/
        /* Get the Vdata information */
        /*****************************/
        status = VSinquire (vdata_id, &n_records, &interlace_mode, 
                            fieldname_list, &vdata_size, vdata_name);
        if (status == FAIL) h4error ("VSinquire");
        status = VSgetclass (vdata_id, vdata_class_name);
        if (DEBUG) { 
            printf ("\nVdata name = %s, class name = %s\n", 
                    vdata_name, vdata_class_name);
            printf ("\t# of records: %d\n\t"
                    "Interlace mode: %s\n\tFields: %s - %d bytes\n\n", 
                    n_records,
                    interlace_mode == FULL_INTERLACE ? "FULL" : "NONE",
                    fieldname_list, vdata_size);
            fflush (stdout);
        }

        /********************************************************************/
        /* Create a global attribute for this Vdata.  All the Vdata         */
        /* attributes will be captured as entries of the global atrribute.  */
        /********************************************************************/
        strcpy (cdfAttr.name, "Vdata_");
        strcat (cdfAttr.name, vdata_name);
        cdf_get_new_attribute_name (cdfAttr.name, GLOBAL_SCOPE,
                                    &create_new_attr, &new_attr_name_created);
        if (create_new_attr) {
            cdfAttr.num = cdf_create_attribute (cdfAttr.name, GLOBAL_SCOPE);
            cdfAttr.entryNum = -1;
            cdf_add_gAttribute_entry (&cdfAttr, "Original Vdata name = ", 
                                      vdata_name);
            if (strcmp(vgroup_name,"") != 0) 
                cdf_add_gAttribute_entry (&cdfAttr,
                                     "Vgroup name that contains this Vdata = ",
                                     vgroup_name);

            cdf_add_gAttribute_entry (&cdfAttr, "Fields in Vdata = ", 
                                      fieldname_list);
        }

        /****************************************************************/
        /*  Process annotation objects (AN_DATA_LABEL and AN_DATA_DESC) */
        /*  in this Vdata if there are any.                             */
        /****************************************************************/
        vdata_tag = DFTAG_VS;
        process_AN_DATA_objects (an_id, vdata_tag, vdata_ref, &cdfAttr);

        /********************************/
        /* Process the Vdata attributes */
        /********************************/
        n_vdata_attrs = VSfnattrs (vdata_id, _HDF_VDATA);
        process_vdata_attr (vdata_id, _HDF_VDATA, n_vdata_attrs, "", &cdfAttr);

        /***********************************************/
        /* Process the Vdata field data and attributes */
        /***********************************************/
        n_fields = VFnfields (vdata_id);
        if (n_fields <= 0) return;

        /********************************************************************/
        /* If this Vdata belongs to a Vgroup, then capture the Vgroup name. */
        /********************************************************************/
        lone_vdata = TRUE;
        if (strcmp(vgroup_name,"") != 0) {
            lone_vdata = FALSE;
            strcpy (vgroup_attr_name, "Vgroup_name");
            cdf_get_new_attribute_name (vgroup_attr_name, VARIABLE_SCOPE,
                                        &create_new_attr, 
                                        &new_attr_name_created);
            if (create_new_attr)
                vgroup_attr_num = cdf_create_attribute(vgroup_attr_name,
                                                       VARIABLE_SCOPE);
        }

        strcpy (vdata_attr_name, "Vdata_name");
        cdf_get_new_attribute_name (vdata_attr_name, VARIABLE_SCOPE,
                                    &create_new_attr, &new_attr_name_created);
        if (create_new_attr)
            vdata_attr_num = cdf_create_attribute(vdata_attr_name,
                                                  VARIABLE_SCOPE);

        /*******************************/
        /* Process the fields in Vdata */
        /*******************************/
        if (DEBUG) {
            printf ("\n  Processing Vdata fields...\n");
            fflush (stdout);
        }

        for (index=0; index < n_fields; index++) {
             strcpy(field.name, VFfieldname (vdata_id, index));
             field.datatype = VFfieldtype (vdata_id, index);
             field.numValues = VFfieldorder (vdata_id, index); 
             if (DEBUG) {
                 printf ("\t%d) %s: %s, order/#of values = %d\n",
                         index, field.name, hdf4_datatype(field.datatype), 
                         field.numValues); 
                 fflush (stdout);
              }


             /*************************/
             /* Create a CDF variable */
             /*************************/
             strcpy (cdfVar.name, field.name);
             cdf_get_new_variable_name (cdfVar.name, &var_already_exists);

             if (field.numValues <= 1) {    /* # of values on each record */
                 if (field.datatype == DFNT_CHAR8)
                     hdf4_dimensionality = 1;
                 else
                     hdf4_dimensionality = 0;
                 hdf4_dim_sizes[0] = 1;   
             }
             else {
                 hdf4_dimensionality = 1;
                 hdf4_dim_sizes[0] = field.numValues;
             }

             record_variable = FALSE;   
             if (n_records > 1)
                 cdfVar.recVariance = TRUE;
             else
                 cdfVar.recVariance = FALSE;

             cdfVar.num = cdf_create_var (&cdfVar, field.datatype, 
                                          hdf4_dimensionality, 
                                          hdf4_dim_sizes, 
                                          record_variable);

             if (cdfVar.num == UNKNOWN_DATATYPE) {
                 printf ("Unknown VD HDF4 data type: %d\n", field.datatype);
                 return;         /* skip the current data set */
             }
             if (COMPRESS) {
               if (cdf_setCompression (cdfVar, n_records) != 1) {
                 /* optional - to eliminate wasted space */
                 if (n_records > 0) 
                   cdf_allocate_zvar_records (cdfVar.num, (long) n_records); 
               }
             } else
               if (n_records > 0) 
                 cdf_allocate_zvar_records (cdfVar.num, (long) n_records);

             if (!lone_vdata)
                 cdf_create_vAttribute_entry (vgroup_attr_name,
                                              cdfVar.name,
                                              CDF_CHAR,
                                              strlen(vgroup_name),
                                              vgroup_name);

             cdf_create_vAttribute_entry (vdata_attr_name,
                                          cdfVar.name,
                                          CDF_CHAR,
                                          strlen(vdata_name), 
                                          vdata_name);

             /**************************************************************/
             /* Create the variable attribute and its entry if one is      */
             /* defined in the mapping file.                               */
             /**************************************************************/
             cdf_handle_attrs_in_mapping_file (cdfVar.name);

             /**************************************************************/
             /* If the variable was already created, this variable name is */
             /* renamed to make sure its name is unique.  Capture the      */
             /* original variable name as an attribute of this variable.   */
             /**************************************************************/
             if (var_already_exists) {
                 strcpy (cdfAttr.name, "Original_variable_name");
                 cdf_get_new_attribute_name (cdfAttr.name, VARIABLE_SCOPE,
                                             &create_new_attr, 
                                             &new_attr_name_created);
                 if (create_new_attr)
                     cdfAttr.num = cdf_create_attribute (cdfAttr.name, 
                                                         VARIABLE_SCOPE);
                 cdf_create_vAttribute_entry ("Original_variable_name",
                                              cdfVar.name,
                                              CDF_CHAR,
                                              strlen(field.name),
                                              field.name);
             }

             /**************************************/
             /* Process the Vdata field attributes */
             /**************************************/
             n_vdata_field_attrs = VSfnattrs (vdata_id, index);
             process_vdata_attr (vdata_id, index, n_vdata_field_attrs,
                                 cdfVar.name, &cdfAttr);

             /***************************************************************/
             /* Process the Vdata field data.  Every field in Vdata has the */
             /* same number of records.                                     */
             /***************************************************************/
             if (n_records >= 1) {
                 allocate_memory (field.datatype, field.numValues, n_records);
                 /*************************************************************
                  * Set the read pointer to the position of the first record    
                  * This MUST be done since the VSread operation is sequential.
                  * Let's assume that a Vdata conatins 3 records and 5 fields.
                  * After the first field is read, the read pointer will be  
                  * pointing the end of the record (4th records) instead of 
                  * the beginning of the record for the next field.        
                  ************************************************************/
                 VSseek (vdata_id, 0);
                 read_vdata_data (vdata_id, field.datatype, n_records, 
                                  field.numValues, field.name, cdfVar);

             }
             if (DEBUG) {
                 printf ("\n");
                 fflush (stdout);
             }
        } 
    }

    status = VSdetach (vdata_id);
}


void get_Vgroup_IDs (int32 file_id)
{
    int32  vgroup_ref;

    /***************************************************************/
    /* Set the reference number to -1 to start the search from the */
    /* beginning of the file.                                      */
    /***************************************************************/
    vgroup_ref = -1;

    /***************************************/
    /* Process all the Vgroups in the file */
    /***************************************/
    while ((vgroup_ref = Vgetid(file_id, vgroup_ref)) != FAIL) {
       vgroup_ids[n_Vgroup_IDs++] = Vattach (file_id, vgroup_ref, "r");
    }
}


/*************************************************************************/
/*  Process the given Vdata index attribute.  The Vdata index can be     */
/*  _HDF_VDATA (vdata), or Vdata field index.                            */
/*************************************************************************/
void process_vdata_attr (int32 vdata_id, int32 field_index, int32 n_attrs,
                         char *cdf_var_name, struct cdfattr *cdfAttr)
{
    intn   status, attr_index;
    int32  index, n_records;
    int    create_new_attr, new_attr_name_created;
    long   cdf_attr_num, cdf_attr_scope;
    char   target[30];
    struct VD_attr   attr;

    if (field_index == _HDF_VDATA) {
        strcpy (target, "VDATA");
        cdf_attr_scope = GLOBAL_SCOPE;
    }
    else {
        strcpy (target, "VDATA FIELD");
        cdf_attr_scope = VARIABLE_SCOPE;
    }

    /***********************************************************************/
    /* Number of records for an attribute is always 1.  There's no concept */
    /* of attribute entry number like CDF in HDF4.  Note that a global CDF */
    /* attribute can have one or more attribute entries.                   */
    /***********************************************************************/
    n_records = 1;

    /***********************************************************************/
    /* Process all the attributes for the given field index.  If the value */
    /* of the field index is _HDF_VDATA, all the attributes asscociated    */
    /* with field index is Vdata attributes.  Otherwise, it's a regular    */
    /* Vdata field attribute.                                              */
    /***********************************************************************/
    for (attr_index=0; attr_index < n_attrs; attr_index++) {
         status = VSattrinfo (vdata_id,
                              field_index,        /* _HDF_VDATA = vdata */ 
                              attr_index,
                              &attr.name[0],
                              &attr.datatype,
                              &attr.n_values,
                              &attr.size);
         if (status == FAIL) h4error ("VSattrinfo - process_vdata_attr");
         if (DEBUG) {
             printf ("  %s attr name = %s, datatype = %s, n_values= %d\n",
                 target,attr.name, hdf4_datatype(attr.datatype), attr.n_values);
             fflush (stdout);
         }
         if (cdf_attr_scope == VARIABLE_SCOPE) {
             cdf_get_new_attribute_name (attr.name, cdf_attr_scope,
                                         &create_new_attr,
                                         &new_attr_name_created);
             if (create_new_attr) 
                 cdfAttr->num = cdf_create_attribute (attr.name,
                                                      cdf_attr_scope); 
         }
         cdfAttr->datatype = get_cdf_datatype (attr.datatype);
         if (cdfAttr->datatype == UNKNOWN_DATATYPE) {
             printf ("Unknown HDF4 data type (process_vdata_attr): %d\n", 
                     attr.datatype);
             continue;         /* skip the current attribute */
         }
         allocate_memory (attr.datatype, attr.n_values, n_records);
         cdfAttr->entryNum++;
         read_vdata_attr_data (vdata_id, field_index, attr_index, 
                               attr, cdf_var_name, cdfAttr); 
         if (DEBUG) printf ("\n"); 
    }
}

/*****************************************************************************/
/*  Process the Vgroup attributes that are attached to the supplied Vgroup.  */
/*****************************************************************************/
void process_vgroup_attr (int32 vgroup_id, struct cdfattr *cdfAttr)
{
    intn   status, n_attrs, attr_index;
    int32  n_records;
    struct VD_attr   attr;

    /***********************************************************************/
    /* Number of records for an attribute is always 1.  There's no concept */
    /* of attribute entry number like CDF in HDF4.  Note that a global CDF */
    /* attribute can have one or more attribute entries.                   */
    /***********************************************************************/
    n_records = 1;

    /***********************************************************************/
    /* Get the number of attributes that belong to this Vgroup and process */
    /* one attribute at a time.                                            */
    /***********************************************************************/
    n_attrs = Vnattrs (vgroup_id);
    if (n_attrs == FAIL) h4error ("Vnattrs - process_vgroup_attr");
    if (DEBUG) {
        printf ("  # of Vgroup attributes = %d\n", n_attrs);
       fflush (stdout);
    }

    for (attr_index=0; attr_index < n_attrs; attr_index++) {
         status = Vattrinfo (vgroup_id,
                             attr_index,
                             &attr.name[0],
                             &attr.datatype,
                             &attr.n_values,
                             &attr.size);
         if (status == FAIL) h4error ("Vattrinfo - process_vgroup_attr");
         if (DEBUG) {
             printf ("  Vgroup attr name = %s, datatype = %s, n_values= %d\n",
                 attr.name, hdf4_datatype(attr.datatype), attr.n_values);
             fflush (stdout);
         }

         cdfAttr->datatype = get_cdf_datatype (attr.datatype);
         if (cdfAttr->datatype == UNKNOWN_DATATYPE) {
             printf ("Unknown HDF4 data type (process_vgroup_attr): %d\n", 
                     attr.datatype);
             continue;         /* skip the current attribute */
         }
         allocate_memory (attr.datatype, attr.n_values, n_records);
         cdfAttr->entryNum++;
         read_vgroup_attr_data (vgroup_id, attr_index, attr.datatype,
                                attr.n_values, cdfAttr);
         if (DEBUG) printf ("\n");
    }
}


/************************************************/
/*  Process the SDS file or dataset attributes  */
/************************************************/
void process_SDS_attrs (int32 id,      /* can be SD or SDS id */
                        int32 n_attrs, /* # of attributes for the given 'id' */
                        long  cdf_attr_scope,
                        char *cdfvar_name)      /* CDF variable name */
{
   int32  attr_index, hdf4_datatype, n_values, n_recs;
   intn   status;
   int    create_new_attr, new_attr_name_created;
   char   sds_attr_name[MAX_NC_NAME];
   struct cdfattr  cdfAttr;

   n_recs = 1;            /* Number of records for an attribute entry */ 
   cdfAttr.scope = cdf_attr_scope;
  
   if (DEBUG) printf("In process_SDS_attrs, cdfvar_name= %s**\n",cdfvar_name);

   for (attr_index=0; attr_index < n_attrs; attr_index++) {
        status = SDattrinfo (id, attr_index, sds_attr_name, &hdf4_datatype, 
                             &n_values);
        if (status == FAIL) h4error ("SDattrinfo");
        if (DEBUG) {
            if (cdf_attr_scope == GLOBAL_SCOPE)
                printf ("\n file attr name= %s\n", sds_attr_name);
            else
                printf ("\n   attr name= %s\n", sds_attr_name);
        }

        if (cdfAttr.scope == GLOBAL_SCOPE) {
            strcpy (cdfAttr.name, "SDS_");
            strcat (cdfAttr.name, sds_attr_name);
        }
        else
            strcpy (cdfAttr.name, sds_attr_name);

        /* Ensure that there are no duplicate attributes */
        cdf_get_new_attribute_name (cdfAttr.name, cdfAttr.scope,
                                    &create_new_attr, &new_attr_name_created);

        if (create_new_attr) {
            cdfAttr.num = cdf_create_attribute(cdfAttr.name, cdfAttr.scope);
        }
        allocate_memory (hdf4_datatype,
                         n_values,            /* # of values on a record */
                         n_recs);  

        read_sds_attr_data (hdf4_datatype, id, attr_index, n_values, n_recs,
                            cdfAttr,          /* CDF attribute info */
                            cdfvar_name);        /* CDF variable name  */
   }
}

/******************************************************/
/*  Process the GR file or a raster image attributes  */
/******************************************************/
void process_GR_attrs (int32 id,       /* can be GR or raster image id */
                       int32 n_attrs,  /* # of attributes for the given 'id' */
                       long cdf_attr_scope,
                       char *cdf_variable_name)   
{
   int32  attr_index, hdf4_datatype, n_values, n_recs;
   intn   status;
   int    create_new_attr, new_attr_name_created;
   char   gr_attr_name[H4_MAX_GR_NAME];
   struct cdfattr  cdfAttr;

   n_recs = 1;              /* Number of records for an attribute entry */
   cdfAttr.scope = cdf_attr_scope;

   for (attr_index=0; attr_index < n_attrs; attr_index++) {
        status = GRattrinfo (id, attr_index, gr_attr_name, &hdf4_datatype,
                             &n_values);
        if (status == FAIL) h4error ("GRattrinfo");
        if (DEBUG) {
            if (cdf_attr_scope == GLOBAL_SCOPE)
                printf ("\n file attr name= %s\n", gr_attr_name);
            else       /* VARIABLE_SCOPE */
                printf ("\n   attr name= %s\n", gr_attr_name);
        }

        if (cdf_attr_scope == GLOBAL_SCOPE) {
            strcpy (cdfAttr.name, "GR_");
            strcat (cdfAttr.name, gr_attr_name);
        }
        else
            strcpy (cdfAttr.name, gr_attr_name);

        /*************************************************/
        /* Ensure that there are no duplicate attributes */
        /*************************************************/
        cdf_get_new_attribute_name (cdfAttr.name, cdfAttr.scope, 
                                    &create_new_attr, &new_attr_name_created);
        if (create_new_attr)
            cdfAttr.num = cdf_create_attribute(cdfAttr.name, cdfAttr.scope);

        allocate_memory (hdf4_datatype,
                         n_values,            /* # of values on a record */
                         n_recs);

        read_gr_attr_data (hdf4_datatype, id, attr_index, n_values, n_recs,
                           cdfAttr, cdf_variable_name);     
   }
}


/********************************************************************/
/*  Process the AN_DATA_LABEL and AN_DATA_DESC annotation objects   */
/*  for the given object reference (obj_ref).                       */
/********************************************************************/
void process_AN_DATA_objects (int32  an_id,    /* Annotation interface ID   */
                              uint16 obj_tag,
                              int32  obj_ref,
                              struct cdfattr  *cdfAttr)  
{
    if (DEBUG) printf ("  tag = %d, ref = %d\n", obj_tag, obj_ref);

    read_data_annotation (an_id, AN_DATA_LABEL, obj_tag, (uint16) obj_ref,
                          cdfAttr);

    read_data_annotation (an_id, AN_DATA_DESC, obj_tag, (uint16) obj_ref,
                          cdfAttr);
}


/************************************************************************/
/*  Read the annotations that are of type 'annotation_type' and belong  */
/*  to the object identified by obj_tag (object tag) and obj_ref        */
/*  (object reference number).                                          */
/************************************************************************/
void read_data_annotation (int32 an_id, ann_type annotation_type,
                           uint16 obj_tag, uint16 obj_ref,
                           struct cdfattr  *cdfAttr)
{
    intn   n_data_labels_or_descs, status;       /* SUCCEED or FAIL */
    int32  *annotation_ids, index;
    char   cdf_var_name[CDF_VAR_NAME_LEN256+1];

    /*************************************************/
    /* Get the number of data labels or descriptions */
    /*************************************************/
    n_data_labels_or_descs = ANnumann (an_id, annotation_type,obj_tag,obj_ref);
    if (n_data_labels_or_descs == FAIL) {
        if (annotation_type == AN_DATA_LABEL)
            h4error ("ANnumann - AN_DATA_LABEL");
        else
            h4error ("ANnumann - AN_DATA_DESCS");
    }
    if (DEBUG) 
        printf ("  # of %s = %d\n", str_annotation_type(annotation_type),
                n_data_labels_or_descs);

    if (n_data_labels_or_descs > 0) {
        annotation_ids = malloc (sizeof(int32) * n_data_labels_or_descs);
        /*********************************************************************/
        /* Get the annotation identifiers that are of type 'annotation_type' */
        /* and belong to the object identified by obj_tag (object tag) and   */
        /* obj_ref (object reference number).                                */
        /*********************************************************************/
        status = ANannlist (an_id, annotation_type, obj_tag, obj_ref,
                            annotation_ids);

        /*******************************************************************/
        /* If the value of the CDF variable name is empty, then CDF global */
        /* attributes are created for AN_DATA_LABELS or AN_DATA_DESCS.     */
        /*******************************************************************/
        strcpy (cdf_var_name, "");

        /*******************************/
        /* Process all the annotations */
        /*******************************/
        for (index=0; index < n_data_labels_or_descs; index++) {
             if (DEBUG) 
                 printf ("%s index %d:\n", 
                         str_annotation_type(annotation_type), index);

             cdfAttr->entryNum++;
             read_annotation (an_id, annotation_ids[index], cdfAttr);
        }
        free (annotation_ids);
    }
}


/******************************************************************/
/*  Return the string representation of the given HDF4 data type  */
/******************************************************************/
char *  hdf4_datatype (int32 hdf4_type)
{
    switch (hdf4_type) {
      case DFNT_CHAR8: 
         return "DFNT_CHAR8";     /* 8-bit character */

      case DFNT_UCHAR8:
        return "DFNT_UCHAR8";     /* 8-bit unsigner character */

      case DFNT_INT8:
        return "DFNT_INT8";       /* 8-bit integer */

      case DFNT_UINT8:
        return "DFNT_UINT8";      /* 8-bit unsigned integer */

      case DFNT_INT16:
        return "DFNT_INT16";      /* 16-bit integer */

      case DFNT_UINT16:
        return "DFNT_UINT16";     /* 16-bit unsigned integer */

      case DFNT_INT32:
        return "DFNT_INT32";      /* 32-bit integer */

      case DFNT_UINT32:
        return "DFNT_UINT32";     /* 32-bit unsigned integer */

      case DFNT_FLOAT32:
        return "DFNT_FLOAT32";    /* 32-bit floating point */

      case DFNT_FLOAT64:
        return "DFNT_FLOAT64";    /* 64-bit floating point */

      default:
        return "UNKNOWN";
    }
}


/************************************************************************/
/*  Return the CDF datatype that corresponds to the given HDF4 datatype */
/************************************************************************/
long get_cdf_datatype (int32 hdf4_type)
{
    switch (hdf4_type) {
      case DFNT_CHAR8: 
         return CDF_CHAR;

      case DFNT_INT8:
        return CDF_INT1;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
        return CDF_UINT1;

      case DFNT_INT16:
        return CDF_INT2;

      case DFNT_UINT16:
        return CDF_UINT2;

      case DFNT_INT32:
        return CDF_INT4;

      case DFNT_UINT32:
        return CDF_UINT4;

      case DFNT_FLOAT32:
        return CDF_REAL4;

      case DFNT_FLOAT64:
        return CDF_REAL8;

      default:
        return UNKNOWN_DATATYPE;
    }
}


/**************************************************************************/
/*  This routine prints the given Scientific Data Set (SDS) information,  */
/*  and it is only here for the debugging purpose.                        */
/**************************************************************************/
void print_SDgetinfo(int32 sds_id, char *sds_name, 
                     int32 rank,       /* Dimensionality */
                     int32 *dim_sizes, int32 data_type, int32 n_sds_attrs)
{
    int i;

    printf ("SDS name = %s, rank= %d, dt=%s, nAttrs=%d, dimSizes= ",
            sds_name, rank, hdf4_datatype(data_type), n_sds_attrs);

    for (i=0; i < rank; i++) {
         if (i > 0) printf (",");
         printf ("%d", dim_sizes[i]);
    }
    if (SDisrecord(sds_id)) 
        printf (", VARY");
    else 
        printf (", NOVARY");
    printf("\n");
}


/***************************************************************************/
/*  This routine prints the given raster image information, and it is only */
/*  here for the debugging purpose.                                        */
/***************************************************************************/
void print_GRgetiminfo (int32 raster_id, char *raster_name, 
                        int32 n_pixel_components, int32 data_type, 
                        int32 interlace_mode, int32 *dim_sizes, 
                        int32 n_raster_attrs)
{
    int   i;
    char  image_type[30];

    if (n_pixel_components == 1)
        strcpy (image_type, "8-bit");

    else if (n_pixel_components == 3)
        strcpy (image_type, "24-bit");
    else
        strcpy (image_type, "general raster");

    printf ("Raster name= %s, type= %s, dt=%s, nAttrs=%d, dimSizes= ",
            raster_name, image_type, hdf4_datatype(data_type), n_raster_attrs);
    for (i=0; i < 2; i++) {
         if (i > 0) printf (",");
         printf ("%d", dim_sizes[i]);
    }
    printf("\n");
    printf ("  # of pixel components = %d\n", n_pixel_components); 
}


/***************************************/
/*  Print the given raster image data  */
/***************************************/
void print_raster_image_data (char *raster_image_type, long n_pixels, 
                              uint8 *raster_data)
{
   long  i;

   printf ("%s\n", raster_image_type);
   for (i=0; i < n_pixels; i++) {
        if (i > 0) printf (" ");
        printf ("%d", *(raster_data+i));
   }
   printf("\n");
}


/****************************************************************/
/*  This routine allocates memory for the given HDF4 datatype.  */
/****************************************************************/
void allocate_memory (int32 hdf4_type, 
                      int32 numElements,       /* # of values on each record */
                      int32 numRecs)           /* # of records to read       */
{
    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         charPtr = malloc (sizeof(char) * numElements * numRecs + 1);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         ucharPtr = malloc (sizeof(unsigned char) * numElements * numRecs);
         break;

      case DFNT_INT16:
         shortPtr = malloc (sizeof(short) * numElements * numRecs);
         break;

      case DFNT_UINT16:
         ushortPtr = malloc (sizeof(unsigned short) * numElements * numRecs);
         break;

      case DFNT_INT32:
         longPtr = malloc (sizeof(long) * numElements * numRecs);
         break;

      case DFNT_UINT32:
         ulongPtr = malloc (sizeof(unsigned long) * numElements * numRecs);
         break;

      case DFNT_FLOAT32:
         floatPtr = malloc (sizeof(float) * numElements * numRecs);
         break;

      case DFNT_FLOAT64:
         doublePtr = malloc (sizeof(double) * numElements * numRecs);
         break;

      default:
         break;
    }
}


/***********************************************************/
/*  This routine reads the SDS data for the given SDS ID.  */
/***********************************************************/
void read_sds_data (int32 sds_id, char *sds_name, int32 hdf4_type, int32 rank, 
                    int32 *dim_sizes, int32 numElements, int32 numRecs,
                    struct cdfvar  cdfVar)
{
    int   i;
    int32 start[CDF_MAX_DIMS], edges[CDF_MAX_DIMS];
    intn  status;

    if (DEBUG) printf ("\n");

    for (i=0; i < rank; i++) {
         start[i] = 0;
         edges[i] = dim_sizes[i];
    }

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) charPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_CHAR8");
         *(charPtr+(numElements*numRecs)) = '\0';
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            charPtr);              /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) ucharPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_UCHAR8");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ucharPtr);       /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) shortPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_INT16");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            shortPtr);       /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) ushortPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_UINT16");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ushortPtr);       /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) longPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_INT32");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            longPtr);       /* data to be written     */
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) ulongPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_UINT32");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ulongPtr);       /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) floatPtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_FLOAT32");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            floatPtr);       /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = SDreaddata (sds_id, start, NULL, edges, (VOIDP) doublePtr);
         if (status == FAIL) h4error("SDreaddata - DFNT_FLOAT64");
         cdf_write_vardata (cdfVar.num,
                            0L,             /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            doublePtr);       /* data to be written     */ 
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         free (doublePtr);
         break;

      default:
         break;
    }
}


/***********************************************************************/
/*  This routine reads the raster image data for the given raster ID.  */
/***********************************************************************/
void read_raster_data (int32 raster_id, int32 hdf4_type, int32 interlace_mode,
                       int32 *dim_sizes, int32 numElements, int32 numRecs,
                       struct cdfvar  cdfVar)
{
    int32 start[CDF_MAX_DIMS], edges[CDF_MAX_DIMS];
    intn  status;

    if (DEBUG) printf ("\n");

    /**********************************/
    /* Define where to read data from */
    /**********************************/
    start[0] = 0;
    start[1] = 0;

    /*************************************************************************/
    /* IMPORTANT NOTE:                                                       */
    /*   Although the underlying data is stored in row-major order in both   */
    /*   SDS and GR interfaces, the SDreaddata and GRreadimage routines      */
    /*   expects the dimension sizes as following:                           */ 
    /*                                                                       */
    /*   - For SDreaddata, dimension sizes are expressed in [row, col] order */
    /*   - For GRreadimage, dimension sizes are expressed in [col,row] order */
    /*                                                                       */
    /* Although the HDF4 documention says the above, GRreadimage expects the */
    /* dimension sizes in [row, column] order just like the SDreaddata       */
    /* routine.  Have a feeling that the documentation is outdated.          */
    /*                                                                       */
    /* dim_sizes[0] contains the number of rows in the image                 */
    /* dim_sizes[1] contains the number of columns in the image              */
    /*************************************************************************/
    edges[0] = dim_sizes[0];      /* # of columns to read */
    edges[1] = dim_sizes[1];      /* # of rows to read    */

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP) charPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_CHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         *(charPtr+(numElements*numRecs)) = '\0';
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            charPtr);              /* data to be written     */
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP) ucharPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_UCHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ucharPtr);             /* data to be written     */
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP) shortPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            shortPtr);             /* data to be written     */
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP)ushortPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ushortPtr);            /* data to be written     */
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP) longPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_INT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            longPtr);              /* data to be written     */
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP) ulongPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ulongPtr);             /* data to be written     */
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP) floatPtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            floatPtr);             /* data to be written     */
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = GRreadimage (raster_id, start, NULL, edges, (VOIDP)doublePtr);
         if (status == FAIL) h4error("GRreadimage - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            doublePtr);            /* data to be written     */
         free (doublePtr);
         break;

      default:
         break;
    }
}


/*******************************************************************/
/*  This routine reads the palette data for the given palette ID.  */
/*******************************************************************/
void read_palette_data (int32 palette_id, int32 hdf4_type, 
                        int32 numElements,       /* # of entries in palette */
                        int32 numRecs,
                        char  *cdf_var_name,
                        struct cdfattr  *cdfAttr)
{
    intn  status;

    if (DEBUG) printf ("\n");

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = GRreadlut (palette_id, (VOIDP) charPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_CHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      charPtr);
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = GRreadlut (palette_id, (VOIDP) ucharPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_UCHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      ucharPtr);
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = GRreadlut (palette_id, (VOIDP) shortPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      shortPtr);
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = GRreadlut (palette_id, (VOIDP)ushortPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      ushortPtr);
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = GRreadlut (palette_id, (VOIDP) longPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_INT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      longPtr);
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = GRreadlut (palette_id, (VOIDP) ulongPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      ulongPtr);
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = GRreadlut (palette_id, (VOIDP) floatPtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      floatPtr);
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = GRreadlut (palette_id, (VOIDP)doublePtr);
         if (status == FAIL) h4error("GRreadlut - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr->name,
                                      cdf_var_name,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      doublePtr);
         free (doublePtr);
         break;

      default:
         break;
    }
}


/*****************************/
/*  Read SDS attribute data  */ 
/*****************************/
void read_sds_attr_data (int32 hdf4_type,
                         int32 id,                 /* can be SD or SDS id    */
                         int32 attr_index,
                         int32 numElements,        /* # of attribute values  */ 
                         int32 numRecs,            /* # of recs - always 1   */
                         struct cdfattr  cdfAttr,
                         char  *cdfvar_name)       /* CDF variable name */
{
    intn  status;
    long  cdf_attr_datatype;

    cdf_attr_datatype = get_cdf_datatype (hdf4_type);
    if (DEBUG) {
        printf("    HDF4 Data Type = %s\n",hdf4_datatype(hdf4_type));
        printf("    attrName= %s, varName= %s, numElements = %d\n",
               cdfAttr.name, cdfvar_name, numElements);
    }

    /*******************************************************************/
    /* HDF4's global attributes (AN, GR, SDS) can be only characters.  */
    /* That's why there's no call for cdf_create_vAttribute_entry for  */
    /* non-character datatypes.                                        */
    /*******************************************************************/
    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = SDreadattr (id, attr_index, (VOIDP) charPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_CHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         *(charPtr+(numElements*numRecs)) = '\0';
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          charPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          charPtr);
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = SDreadattr (id, attr_index, (VOIDP) ucharPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_UCHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          ucharPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          ucharPtr);
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = SDreadattr (id, attr_index, (VOIDP) shortPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          shortPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          shortPtr);
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = SDreadattr (id, attr_index, (VOIDP) ushortPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          ushortPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          ushortPtr);
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = SDreadattr (id, attr_index, (VOIDP) longPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_INT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          longPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          longPtr);
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = SDreadattr (id, attr_index, (VOIDP) ulongPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          ulongPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          ulongPtr);
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = SDreadattr (id, attr_index, (VOIDP) floatPtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          floatPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          floatPtr);
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = SDreadattr (id, attr_index, (VOIDP) doublePtr);
         if (status == FAIL) h4error("SDreadattr - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          doublePtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdfvar_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          doublePtr);
         free (doublePtr);
         break;

      default:
         break;
    }
}


/*****************************/
/*  Read SDS attribute data  */
/*****************************/
void read_gr_attr_data (int32 hdf4_type,
                        int32 id,            /* can be GR or raster image id */
                        int32 attr_index,
                        int32 numElements,         /* # of attribute avalues */
                        int32 numRecs,             /* # of recs - always 1   */
                        struct cdfattr  cdfAttr,
                        char  *cdf_var_name)       /* CDF var name */
{
    intn  status;
    long  cdf_attr_datatype;

    cdf_attr_datatype = get_cdf_datatype (hdf4_type);

    /*******************************************************************/
    /* HDF4's global attributes (AN, GR, SDS) can be only characters.  */
    /* That's why there's no call for cdf_create_vAttribute_entry for  */
    /* non-character datatypes.                                        */
    /*******************************************************************/
    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = GRgetattr (id, attr_index, (VOIDP) charPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_CHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (cdfAttr.scope == GLOBAL_SCOPE)
             cdf_create_gAttribute_entry (cdfAttr.num,
                                          0,              /* entry number */
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          charPtr);
         else   /* VARIABLE_SCOPE */
             cdf_create_vAttribute_entry (cdfAttr.name,
                                          cdf_var_name,
                                          cdf_attr_datatype,
                                          (long) numElements,
                                          charPtr);
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = GRgetattr (id, attr_index, (VOIDP) ucharPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_UCHAR8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      ucharPtr);
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = GRgetattr (id, attr_index, (VOIDP) shortPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      shortPtr);
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = GRgetattr (id, attr_index, (VOIDP) ushortPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      ushortPtr);
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = GRgetattr (id, attr_index, (VOIDP) longPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_INT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      longPtr);
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = GRgetattr (id, attr_index, (VOIDP) ulongPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      ulongPtr);
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = GRgetattr (id, attr_index, (VOIDP) floatPtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      floatPtr);
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = GRgetattr (id, attr_index, (VOIDP) doublePtr);
         if (status == FAIL) h4error("GRgetattr - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_vAttribute_entry (cdfAttr.name,
                                      cdf_var_name,
                                      cdf_attr_datatype,
                                      (long) numElements,
                                      doublePtr);
         free (doublePtr);
         break;

      default:
         break;
    }
}


/*********************/
/*  Read Vdata  data */
/*********************/
void read_vdata_data (int32 vdata_id, int32 hdf4_type, 
                      int32 numRecs,        /* # of records in Vdata */
                      int32 numElements,    /* # of values in a field */
                      char *fieldname,
                      struct cdfvar  cdfVar)
{
    intn  status;

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_CHAR8/DFNT_INT8");
         status = VSread (vdata_id, (uint8 *) charPtr, numRecs, FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_CHAR8/INT8");
         *(charPtr+(numElements*numRecs)) = '\0';
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            charPtr);              /* data to be written     */
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_UCHAR8/DFNT_UINT8");
         status = VSread (vdata_id, (uint8 *)ucharPtr, numRecs, FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_CHAR8/UINT8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ucharPtr);             /* data to be written     */
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_INT16");
         status = VSread (vdata_id, (uint8 *)shortPtr, numRecs, FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            shortPtr);             /* data to be written     */
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_UINT16");
         status = VSread (vdata_id, (uint8 *)ushortPtr, numRecs,FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ushortPtr);             /* data to be written     */
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_INT32");
         if (DEBUG) {
             printf ("  Vdata numRecs = %d\n", numRecs);
             fflush(stdout);
         }
         status = VSread (vdata_id, (uint8 *)longPtr, numRecs, FULL_INTERLACE);
         if (status == FAIL) 
             h4error("VSread - DFNT_INT32");
         else {
             if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
             cdf_write_vardata (cdfVar.num,
                                0L,                /* Starting at rec # 1 */
                                (long) numRecs,    /* # of records to write  */
                                cdfVar.dimensionality,
                                cdfVar.dimSizes,
                                longPtr);          /* data to be written     */
         }
         if (longPtr != NULL) free (longPtr);
         break;

      case DFNT_UINT32:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_UINT32");
         status = VSread (vdata_id, (uint8 *)ulongPtr, numRecs, FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            ulongPtr);              /* data to be written     */
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_FLOAT32");
         status = VSread (vdata_id, (uint8 *)floatPtr, numRecs, FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            floatPtr);             /* data to be written     */
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = VSsetfields (vdata_id, fieldname);
         if (status == FAIL) h4error("VSetfields- DFNT_FLOAT64");
         status = VSread (vdata_id, (uint8 *)doublePtr, numRecs,FULL_INTERLACE);
         if (status == FAIL) h4error("VSread - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_write_vardata (cdfVar.num,
                            0L,            /* Write data starting at rec # 1 */
                            (long) numRecs,        /* # of records to write  */
                            cdfVar.dimensionality,
                            cdfVar.dimSizes,
                            doublePtr);            /* data to be written     */
         free (doublePtr);
         break;

      default:
         break;
    }
}


/********************************************************/
/*  Read Vdata attribute or Vdata field attribute data  */ 
/********************************************************/
void read_vdata_attr_data (int32 vdata_id, 
                           int32 field_index,
                           intn  attr_index,
                           struct VD_attr  attr,    /* Vdata field structure */
                           char *cdf_var_name,
                           struct cdfattr  *cdfAttr)
{
    intn  status;
    int32 hdf4_type;
    long  numElements, numRecs;

    numRecs     = 1;
    numElements = attr.n_values;
    hdf4_type   = attr.datatype;

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = VSgetattr (vdata_id, field_index, attr_index, (VOIDP)charPtr);
         if (status == FAIL) h4error("VSgetattr- DFNT_CHAR8/DFNT_INT8");
         *(charPtr+(numElements*numRecs)) = '\0';
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          charPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          charPtr); 
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP)ucharPtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_UCHAR8/UINT8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          ucharPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          ucharPtr); 
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP)shortPtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          shortPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          shortPtr); 
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP)ushortPtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          ushortPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          ushortPtr); 
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP) longPtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_INT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          longPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          longPtr); 
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP)ulongPtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          ulongPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          ulongPtr); 
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP)floatPtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          floatPtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          floatPtr); 
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = VSgetattr (vdata_id,field_index,attr_index, (VOIDP)doublePtr);
         if (status == FAIL) h4error("VSgetattr - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         if (field_index == _HDF_VDATA)         /* Vdata attribute */
             cdf_create_gAttribute_entry (cdfAttr->num,
                                          cdfAttr->entryNum, 
                                          cdfAttr->datatype,
                                          (long) numElements,
                                          doublePtr);
         else     /* Vdata field attribute */
             cdf_create_vAttribute_entry (attr.name, 
                                          cdf_var_name, 
                                          cdfAttr->datatype,
                                          numElements, 
                                          doublePtr); 
         free (doublePtr);
         break;

      default:
         break;
    }
}


/********************************************************/
/*  Read Vdata attribute or Vdata field attribute data  */
/********************************************************/
void read_vgroup_attr_data (int32 vgroup_id,
                            intn  attr_index,
                            int32 hdf4_type,
                            int32 numElements,
                            struct cdfattr  *cdfAttr)
{
    intn  status;
    int32 numRecs;

    numRecs     = 1;

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) charPtr);
         if (status == FAIL) h4error("Vgetattr- DFNT_CHAR8/DFNT_INT8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      charPtr);
         free (charPtr);
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) ucharPtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_UCHAR8/UINT8");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      ucharPtr);
         free (ucharPtr);
         break;

      case DFNT_INT16:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) shortPtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_INT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      shortPtr);
         free (shortPtr);
         break;

      case DFNT_UINT16:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) ushortPtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_UINT16");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      ushortPtr);
         free (ushortPtr);
         break;

      case DFNT_INT32:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) longPtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_INT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      longPtr);
         free (longPtr);
         break;

      case DFNT_UINT32:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) ulongPtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_UINT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      ulongPtr);
         free (ulongPtr);
         break;

      case DFNT_FLOAT32:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) floatPtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_FLOAT32");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      floatPtr);
         free (floatPtr);
         break;

      case DFNT_FLOAT64:
         status = Vgetattr (vgroup_id, attr_index, (VOIDP) doublePtr);
         if (status == FAIL) h4error("Vgetattr - DFNT_FLOAT64");
         if (DEBUG_DATA) print_hdf4_data (hdf4_type, numElements, numRecs);
         cdf_create_gAttribute_entry (cdfAttr->num,
                                      cdfAttr->entryNum,
                                      cdfAttr->datatype,
                                      (long) numElements,
                                      doublePtr);
         free (doublePtr);
         break;

      default:
         break;
    }
}


/*******************************************************************/
/*  Print the data stored in one of the global pointer variables.  */
/*******************************************************************/
void print_hdf4_data (int32 hdf4_type, 
                      int32 numElements,    /* # of values on each record */
                      int32 numRecs)        /* # of records to print      */
{
    int  i, j;

    switch (hdf4_type) {
      case DFNT_CHAR8:
      case DFNT_INT8:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) {
                   if (j > 0) printf (" ");
                   if (hdf4_type == DFNT_CHAR8)
                       printf ("%c", *(charPtr+(i*numElements)+j)); 
                   else
                       printf ("%d", *(charPtr+(i*numElements)+j)); 
              }
              printf ("\n");
         }
         break;

      case DFNT_UCHAR8:
      case DFNT_UINT8:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) {
                   if (j > 0) printf (" ");
                   printf ("%d", *(ucharPtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      case DFNT_INT16:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) { 
                   if (j > 0) printf (" ");
                   printf ("%d", *(shortPtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      case DFNT_UINT16:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) { 
                   if (j > 0) printf (" ");
                   printf ("%d", *(ushortPtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      case DFNT_INT32:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) { 
                   if (j > 0) printf (" ");
                   printf ("%ld", *(longPtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      case DFNT_UINT32:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) { 
                   if (j > 0) printf (" ");
                   printf ("%ld", *(ulongPtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      case DFNT_FLOAT32:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) { 
                   if (j > 0) printf (" ");
                   printf ("%4.1f", *(floatPtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      case DFNT_FLOAT64:
         for (i=0; i < numRecs; i++) {
              printf ("  Rec # %d: ", i);
              for (j=0; j < numElements; j++) {
                   if (j > 0) printf (" ");
                   printf ("%4.1f", *(doublePtr+(i*numElements)+j));
              }
              printf ("\n");
         }
         break;

      default:
         break;
    }
}

/*************************************************************************
 Get the number of elements/values on each record and number of records 
 for the given data set id (sds_id).
**************************************************************************/
void get_sds_info (int32 sds_id, int32 rank, int32 *dim_sizes, 
                   int32 *numElements, int32 *numRecs)
{
    int i, begin_index;

    if (SDisrecord(sds_id)) {        /* record-varing variable */
        *numRecs = dim_sizes[0];
        begin_index = 1; 
    }
    else {                           /* not record-varying variable */
        *numRecs = 1;
        begin_index = 0; 
    }

    *numElements = 1;
    for (i=begin_index; i < rank; i++)
         *numElements *= dim_sizes[i];
} 


/*************************************************************************
 Read the file file labels, file descriptions, data labels, or data
 descriptions for the given annotation id (an_id).
**************************************************************************/
void read_AN_file_annotation (int32 an_id, int32 tot_annotation, 
                              ann_type annotation_type)
{
    intn  status_n;
    int   create_new_attr, new_attr_name_created;
    int32 annotation_id, index;
    struct cdfattr  cdfAttr;

    if (tot_annotation < 1) return;

    if (annotation_type == AN_FILE_LABEL)
        strcpy(cdfAttr.name, "File_Label");
    else     /* AN_FILE_DESC */
        strcpy(cdfAttr.name, "File_Description");

    cdfAttr.scope = GLOBAL_SCOPE;
    cdfAttr.entryNum = 0;

    cdf_get_new_attribute_name (cdfAttr.name, cdfAttr.scope, 
                                &create_new_attr, &new_attr_name_created);
    if (create_new_attr)
         cdfAttr.num = cdf_create_attribute (cdfAttr.name, cdfAttr.scope);

    for (index=0; index < tot_annotation; index++) {
         annotation_id = ANselect (an_id, index, annotation_type);
         if (annotation_id == FAIL) h4error ("ANselect");
         if (DEBUG) printf ("%s index %d:\n",
                            str_annotation_type(annotation_type), index);

         read_annotation (an_id, annotation_id, &cdfAttr);
         cdfAttr.entryNum++;
         
         status_n = ANendaccess (annotation_id);
    }
}


/*************************************************************************
 Read the annotation for the given annotation ID. 
**************************************************************************/
void read_annotation (int32 an_id, int32 annotation_id, 
                      struct cdfattr  *cdfAttr)
{
    int32 annotation_length, status32;
    char  *buffer;

    annotation_length = ANannlen (annotation_id);
    buffer = malloc (sizeof(char) * (annotation_length+1));
    status32 = ANreadann (annotation_id, buffer, annotation_length+1);

    /***************************************************************/
    /*  The ANreadann routine doesn't put a NUll character at the  */
    /*  end of the annotation string.  So put a NULL character     */
    /*  at the end of the annotation string.                       */
    /***************************************************************/
    *(buffer+annotation_length) = '\0';

    if (DEBUG) 
         printf ("Annotation length = %d bytes\n%s\n", (int) strlen(buffer), buffer);

    /**************************************/
    /* Create an CDF attribute entry/data */
    /**************************************/
    if (cdfAttr->scope == GLOBAL_SCOPE)
        cdf_create_gAttribute_entry (cdfAttr->num, 
                                     cdfAttr->entryNum,
                                     CDF_CHAR,       /* entry datatype */
                                     (long) strlen(buffer),
                                     buffer);
    // else
    //     cdf_create_vAttribute_entry (cdf_attr_name, cdf_var_name, CDF_CHAR,
    //                                  (long) strlen(buffer),
    //                                  buffer);
 
    free (buffer);
}


/**************************************************************************/
/*  Return the string respresentation of the given HDF4 annotation type.  */
/**************************************************************************/
char * str_annotation_type (ann_type annotation_type)
{
    switch (annotation_type) {
       case AN_FILE_LABEL:
          return "AN_FILE_LABEL"; 
       
       case AN_FILE_DESC:
          return "AN_FILE_DESC"; 

       case AN_DATA_LABEL:
          return "AN_DATA_LABEL"; 
       
       case AN_DATA_DESC:
          return "AN_DATA_DESC"; 

       default:
          return "UNKNOWN";
    }
}
      

/******************************************************************/
/*  Get the latest error code from the error stack and print the  */
/*  error message that corresponds to the error code.             */
/******************************************************************/
void h4error(char *source)
{
   intn error_code = HEvalue (1);    /* Get the latest error code */
   printf ("\n** Error: %s\n\tError code %d - %s\n", 
           source, error_code, HEstring(error_code));
} 


/****************************************************************/
/*  Print the usage if no input is given from the command line  */
/****************************************************************/
void usage()
{
   printf ("\nDescription:\n");
   printf ("    This program converts a HDF4 file into a CDF file.\n");
   printf ("\n");
   printf ("Usage: hdf4-to-cdf -[Options] <HDF4 file name>\n");
   printf ("\n");
   printf ("Options:\n");
   printf ("    -map <map file>    Use the user-supplied mapping file.\n");
   printf ("    -o <output file>   Use the user-supplied output file name.\n");
   printf ("    -debug             Show debug information.\n");
   printf ("    -v                 Show converter, CDF and HDF versions.\n");
   printf ("    -nocompress        Do not compress variable(s). Default is\n");
   printf ("                       to compress.\n");
   printf ("\n");
   printf ("Examples of usage: \n");
   printf ("1) hdf4-to-cdf test.hdf\n");
   printf ("   will convert test.hdf into test.cdf.\n\n");
   printf ("2) hdf4-to-cdf -map mymap.dat test.hdf\n");
   printf ("   will convert the test.hdf into test.cdf using the\n");
   printf ("   user-supplied mapping file called 'mymap.dat' instead\n");
   printf ("   of using the default hdf4-to-CDF mapping file.\n");
   printf ("3) hdf4-to-cdf -o /home/mydir/test /remote_loc/test.hdf\n");
   printf ("   will convert test.hdf into test.cdf.\n\n");
   printf ("4) hdf4-to-cdf -o test1.cdf test.hdf\n");
   printf ("   will convert test.hdf into test1.cdf.\n\n");
   printf ("\n");
   exit(0);
}

/***********************************/
/*  Parse the command line input.  */
/***********************************/
void parseCmdLine (int argc, char *argv[])
{
   int  i;
   char *nextArg;

   for (i=0; i < argc; i++) {
        if (strcmp(argv[i],"-map") == 0) {
            nextArg = getNextArgument(argc, argv, i);
            if (strcmp(nextArg, nomsg) == 0 ||
                *(nextArg+0) == '-')
                errorMsg("** Error - mapping file is not defined");
            else
                 strcpy(mappingFile, nextArg);
        }
        else if (strcmp(argv[i],"-o") == 0) { 
            nextArg = getNextArgument(argc, argv, i);
            if (strcmp(nextArg, nomsg) == 0 ||
                *(nextArg+0) == '-')
                errorMsg("** Error - output file name is not defined");
            else
                 strcpy(output_filename, nextArg);
        } 
        else if (strcmp(argv[i],"-debug") == 0) { 
            DEBUG = TRUE;
            DEBUG_DATA = FALSE;
        } 
        else if (strcmp(argv[i],"-debugall") == 0) {
            DEBUG = TRUE;
            DEBUG_DATA = TRUE;
        }
        else if (strcmp(argv[i],"-v") == 0) { 
            showVersions();
            exit(1);
        } 
        else if (strcmp(argv[i],"-nocompress") == 0) {
            COMPRESS = FALSE;
        }
   }
}


/**************************************************************************/
/*  Called by parseCmdLIne.  This routine retrives the next command line  */
/*  argument.                                                             */
/**************************************************************************/
char * getNextArgument (int argc, char *argv[], int currentIndex)
{
    char *nextArg, msg[30];
    int  nextIndex;

    nextIndex  = currentIndex+1;
    if (nextIndex == argc) {
        nextArg = nomsg;
    }
    else
        nextArg = argv[nextIndex];

    return nextArg;
}


void errorMsg (char *msg)
{
   printf ("%s\n", msg);
   exit (1);
}


/***********************************************/
/*  Load the HDF4-to-CDF mapping information */
/***********************************************/
void load_mapping_info ()
{
    FILE  *inFile;                 /* Pointer to the HDF4-CDF mapping file */
    char  rec[MAX_REC_LENGTH],     /* HDF4-to-CDF mapping record           */
          *ptr;

    if ((inFile = fopen(mappingFile, "r")) == NULL) {
         printf ("**  Cannot open file: %s  **\n", mappingFile);
        exit (1);
    }

    n_mapping_attrs = 0;
    while (fgets(rec, MAX_REC_LENGTH, inFile) != NULL) {
        rec[strlen(rec)-1] = '\0';

         /* Process the mapping record if it's not a comment */
        if (rec[0] != '#'  &&  rec[0] != ' ' && strlen(rec) > 0) {
            sscanf(rec, "%s %s",
                   hdf4AttrNames[n_mapping_attrs],
                   cdfAttrNames[n_mapping_attrs]);
            ptr = (char *) strstr(rec,"//");
            if (ptr != NULL) {
                ptr += 3;
                strcpy(comments[n_mapping_attrs], ptr);
            }

            if (DEBUG)
                printf("%s %s %s\n",
                       hdf4AttrNames[n_mapping_attrs],
                       cdfAttrNames[n_mapping_attrs],
                       comments[n_mapping_attrs]);

            n_mapping_attrs++;
        }
    }
}

/*--------------------------------------------------------------------------
 *  Handles a CDF error.
 *--------------------------------------------------------------------------*/
void cdf_status_handler (CDFstatus status, char *source)
{
   char  message[CDF_STATUSTEXT_LEN+1];

   CDFerror (status, message);              /* Get the appropriate message */

   if (status < CDF_WARN) {
       printf ("An error has occurred, halting...\n");
       printf ("%s\n", message);
       printf ("** Error source: %s\n", source);
       exit (status);
   }
   else if (status < CDF_OK) {
       printf ("Warning, function may not have compeleted as expected...\n");
       printf ("%s\n", message);
   }
   else if (status > CDF_OK) {
       printf ("Function compeleted successfully, but be advised that...\n");
       printf ("%s\n", message);
   }
}


/*--------------------------------------------------------------------------
 *  Map the given HDF4 variable name to an appropriate CDF variable name.
 *--------------------------------------------------------------------------*/
void map_HDF4_name_to_CDF (char *searchName,      /* in - HDF4F var name  */
                           char *mappedName)      /* out - CDF var name   */
{
   int i;

   strcpy(mappedName, searchName);

   for (i=0; i < n_mapping_attrs; i++) {
        if (strcmp(hdf4AttrNames[i], searchName) == 0) {
            strcpy(mappedName, cdfAttrNames[i]);
            break;
        }
   }
}


/*--------------------------------------------------------------------------
 *  Remove the filepath from a file name. *
 *  Example:
 *      /home/data/mydata.txt => mydata.txt
 *--------------------------------------------------------------------------*/
void removeFilepath (char *fileName) {
    char *ptr;

    ptr = strrchr(fileName, '/');        /* Unix file path separator */
    if (ptr != NULL) {
        ptr++;
        strcpy(fileName, ptr);
    }
    else {
       ptr = strrchr(fileName, '\\');     /* Windows file separator */
       if (ptr != NULL) {
           ptr++;
           strcpy(fileName, ptr);
       }
    }
}


/*--------------------------------------------------------------------------
 *  Remove the file extension from the given file name.
 *
 *  Example:
 *      mydata.txt => mydata
 *--------------------------------------------------------------------------*/
char * removeFileExtension (char *fileName, char *fileExtension) {
    char *ptr;

    ptr = (char *) strlaststr(fileName, fileExtension);
    if (ptr != NULL) *ptr = '\0';

   return ptr;
}


/*-----------------------------------------------------------------------------
 *  TRUE if s1 ends with s2.  Otherwise, FALSE is returned.
 *----------------------------------------------------------------------------*/
int endsWith (char *s1, char *s2)
{
    int  i;
    char *ps1, *ps2;

    if (strlen(s2) > strlen(s1))
        return FALSE;

    ps1 = s1 + strlen(s1) - strlen(s2);
    ps2 = s2;

    for (i=0; i < strlen(s2); i++)
        if (*(ps1++) != *(ps2++))
            return FALSE;

    return TRUE;
}


/*-----------------------------------------------------------------------------
 *  Find the last occurence of s2 in s1.  If s21 is not found, s1 is returned.
 *----------------------------------------------------------------------------*/
char * strlaststr (char *s1, char *s2)
{
    char *sc2, *psc1, *ps1;

    if (*s2 == '\0')
        return((char *)s1);

    ps1 = s1 + strlen(s1);

    while(ps1 != s1) {
        --ps1;
        for (psc1 = ps1, sc2 = s2; ; )
                if (*(psc1++) != *(sc2++))
                        break;
                else if (*sc2 == '\0')
                        return ((char *)ps1);
    }
    return ((char *)NULL);
}


/*------------------------------------------------------------------
 *   This routine creates a CDF file.
 *------------------------------------------------------------------*/
void cdf_create_file (char *fileName)
{
    CDFstatus status;

    int   i;
    long  dim,                          /* Array dimensionality      */
          dimSizes[CDF_MAX_DIMS],       /* Array dimension sizes     */
          dimVarys[CDF_MAX_DIMS];       /* Array dimension variances */

    /********************************************************************/
    /*  The following variables need some dummy values for the creation */
    /*  of a CDF file and CDF variables.                                */
    /********************************************************************/
    dim = 0L;
    for (i=0; i < CDF_MAX_DIMS; i++) {
        dimSizes[i] = 1;
        dimVarys[i] = VARY;
    }

    /* The CDF library adds .cdf at the end of the file name */
    status = CDFlib (CREATE_, CDF_, fileName, dim, dimSizes, &id, NULL_);
    if (status != CDF_OK) cdf_status_handler (status, "CREATE_, CDF_");
}


/*--------------------------------------------------------------------------
 *  This routine creates a CDF global or variable attribute.
 *-------------------------------------------------------------------------*/
long cdf_create_attribute (char *attrName,   /* Attribute name to be created */
                           long attrScope)   /* Attribute scope */
{
   CDFstatus status;
   long      attrNum;
   char      msg[80];

   status = CDFlib (CREATE_, ATTR_, attrName,      /* in - attribute name    */
                                    attrScope,     /* in - attribute scope   */
                                    &attrNum,      /* out - attribute number */
                     NULL_);
   if (status != CDF_OK)  {
       strcpy(msg, "CREATE_, ATTR_, ");
       strcat(msg, attrName);
       cdf_status_handler (status, msg);
   }

   strcpy (processed_cdf_attrs[n_cdf_attrs_created].name, attrName);
   processed_cdf_attrs[n_cdf_attrs_created].seq_no = 0;
   processed_cdf_attrs[n_cdf_attrs_created++].scope = attrScope;

   return attrNum;
}


/*--------------------------------------------------------------------------
 *  This routine checks whether the given variable name has already been
 *  created (since CDF doesn't allow duplicate variable names - every variable
 *  name must be unique).  If the given variable name has been already
 *  created, it increments the sequence number by 1 and tack the sequence 
 *  number at the end of the given variable name and return that name. 
 *--------------------------------------------------------------------------*/
void cdf_get_new_variable_name (char *varName, int *var_exists)
{
   int   i, seq_no;
   char  mapped_name[CDF_VAR_NAME_LEN256+1];

   /********************************************************/
   /* Get the user-define mapping name if one is available */
   /********************************************************/
   if (DEBUG) printf ("   In get_new_var_name, orig var name = %s\n", varName);

   map_HDF4_name_to_CDF (varName, mapped_name);

   if (DEBUG) printf ("   In get_new_var_name, mapped name= %s\n", mapped_name);
   strcpy (varName, mapped_name);

   *var_exists = FALSE;
               
   for (i=0; i < n_cdf_vars_created; i++) {
        if (strcmp(processed_cdf_vars[i].name, varName) == 0) {
            *var_exists = TRUE;
            processed_cdf_vars[i].seq_no++;
            seq_no = processed_cdf_vars[i].seq_no;
            break;
        }
   }
   if (*var_exists) {
       sprintf (varName, "%s_%d", varName, seq_no);
       if (DEBUG) printf("  %% new varName = %s\n", varName);
   }
}


/*--------------------------------------------------------------------------
 *  This routine checks whether the given variable name has already been
 *  created (since CDF doesn't allow duplicate variable names - every variable
 *  name must be unique).  If the given variable name has been already
 *  created, it increments the sequence number by 1 and tack the sequence
 *  number at the end of the given variable name and return that name.
 *--------------------------------------------------------------------------*/
void cdf_get_new_attribute_name (char *attrName, long attr_scope,
                                 int *create_new_attr, 
                                 int *new_attr_name_created)
{
   int   i, seq_no, attr_exists;
   char  mapped_name[CDF_ATTR_NAME_LEN256+1];

   /********************************************************/
   /* Get the user-define mapping name if one is available */
   /********************************************************/
   map_HDF4_name_to_CDF (attrName, mapped_name);
   strcpy (attrName, mapped_name);

   attr_exists     = FALSE;
   *create_new_attr = FALSE;
   *new_attr_name_created = FALSE;

   for (i=0; i < n_cdf_attrs_created; i++) {
        if (strcmp(processed_cdf_attrs[i].name, attrName) == 0) {
            attr_exists = TRUE;
            if (attr_scope == GLOBAL_SCOPE) {
                *create_new_attr = TRUE;
            }
            else {     /* VARIABLE_SCOPE */
               if (attr_scope != processed_cdf_attrs[i].scope) {
                   *create_new_attr = TRUE;
               }
            }

            if (*create_new_attr) {
                processed_cdf_attrs[i].seq_no++;
                seq_no = processed_cdf_attrs[i].seq_no;
            }
            break;
        }
   }

   if (!attr_exists) 
       *create_new_attr = TRUE;
   else {    
       if (*create_new_attr) {
           sprintf (attrName, "%s_%d", attrName, seq_no);
           *new_attr_name_created = TRUE;
       }
   }
}


/*--------------------------------------------------------------------------
 *  This routine creates a CDF global attribute entry.  An attribute
 *  can have one or more entries.
 *--------------------------------------------------------------------------*/
void cdf_create_gAttribute_entry (long attrNum,       /* Attr number          */
                                  long entryNum,      /* Attr entry number    */
                                  long datatype,      /* Attr entry datattype */
                                  long numElements,   /* # of attr entries    */
                                  void *data)         /* Attr entry data      */
{
   CDFstatus status;

   /**************************************************************************
    * If the datatype is a string (i.e. CDF_CHAR), the value of 'numElements'
    * must be the length of the string.  For non-string data, the value of
    * 'numElements' represents the number of values in 'data'.
    **************************************************************************/
   if (datatype == CDF_CHAR) {
       numElements = strlen( (char *) data);
       if (numElements == 0) {                   /* Empty data */
           numElements = 1;
           strcpy((char *) data, " ");
       }
   }

   status = CDFlib (SELECT_, ATTR_, attrNum,
                             gENTRY_, entryNum,
                    PUT_, gENTRY_DATA_, datatype,
                                        numElements,
                                        data,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "PUT_, gENTRY_DATA_");
}


/*--------------------------------------------------------------------------
 *  This routine creates a variable attribute entry for the given
 *  variable name and attribute name.  Unlike global attribute, the
 *  variable attribute can't have more than one entry.
 *--------------------------------------------------------------------------*/
void cdf_create_vAttribute_entry (char *attrName,    /* CDF attribute name   */
                                  char *varName,     /* CDF variable name    */
                                  long datatype,     /* Attr entry datattype */
                                  long numElements,  /* # of attr entries    */
                                  void *data)        /* Attr entry data      */
{
   CDFstatus status;

   /**************************************************************************
    * If the datatype is a string (i.e. CDF_CHAR), the value of 'numElements'
    * must be the length of the string.  For non-string data, the value of
    * 'numElements' represents the number of values in 'data'.
    **************************************************************************/
   if (datatype == CDF_CHAR) {
       numElements = strlen( (char *) data);
       if (numElements == 0) {                   /* Empty data */
           numElements = 1;
           strcpy((char *) data, " ");
       }
   }

   status = CDFlib (SELECT_, ATTR_NAME_, attrName,
                             zENTRY_NAME_, varName,
                    PUT_, zENTRY_DATA_, datatype,
                                        numElements,
                                        data,
                    NULL_);
   if (status != CDF_OK) cdf_status_handler (status, "PUT_, zENTRY_DATA_");
}


/*--------------------------------------------------------------------------*/
/* Create a CDF zVariable.                                                  */
/*--------------------------------------------------------------------------*/
long cdf_create_var (struct cdfvar   *cdfVar,
                     int32 hdf4_type, 
                     int32 hdf4_dimensionality,
                     int32 *hdf4_dim_sizes, 
                     long record_variable)
{
   int    i;
   long   numElements,
          dimVariance[CDF_MAX_DIMS],    /* Dimension variance */
          varNum;                       /* zVariable number */
   char   mapped_name[CDF_VAR_NAME_LEN256+1]; 
   CDFstatus       status;

   /***********************************************************************/
   /* If the HDF4 dataset  is a record variable, the first dimension of   */
   /* the variable is a record number and it needs to be skipped.         */
   /***********************************************************************/
   cdfVar->numValues = 1;

   if (record_variable) {
       if (hdf4_dimensionality > 0)       /* variable is not scalar */
           cdfVar->dimensionality = hdf4_dimensionality - 1;
   }
   else 
       cdfVar->dimensionality = hdf4_dimensionality;

   for (i=0; i < hdf4_dimensionality; i++) {
        dimVariance[i] = VARY;
        cdfVar->numValues *= hdf4_dim_sizes[i];
        if (record_variable) {
            if (i == 0)      /* Ignore the record number */
                cdfVar->numValues /= hdf4_dim_sizes[i];
            else
                cdfVar->dimSizes[i-1] = hdf4_dim_sizes[i];
        }
        else
            cdfVar->dimSizes[i] = hdf4_dim_sizes[i];
   }

   cdfVar->datatype = get_cdf_datatype(hdf4_type);
   if (cdfVar->datatype == UNKNOWN_DATATYPE) {
       printf ("*** ERROR: HDF4 datatype for variable \"%s\" is "
               "unsupported\n", cdfVar->name);
       return (long) UNKNOWN_DATATYPE;
   }

   /**************************************************************/
   /*  In CDF, a string of 200 characters is be defined as       */
   /*    CHAR/200 0:[],  not CHAR/1 1:[200]                      */
   /*                                                            */
   /*  A string is treated as 1-D in HDF4 while it is treated    */
   /*  as 0-D in CDF.                                            */
   /**************************************************************/
   if (hdf4_type == DFNT_CHAR8) {
       numElements = hdf4_dim_sizes[hdf4_dimensionality-1];  /* string length */
       cdfVar->dimensionality--;
   }
   else
       numElements = 1;    /* Non-character variables should have this value */

   if (DEBUG) {
       printf("  %%  # of vars created so far = %d\n", n_cdf_vars_created);
       printf ("  Issued CREATE_, zVAR_\n");
       printf ("  var name = %s, datatype= %s, numElements= %ld, Dim=%ld, DimSizes= [",
               cdfVar->name, cdf_str_datatype(cdfVar->datatype), numElements,
               cdfVar->dimensionality);
       for (i=0; i < cdfVar->dimensionality; i++) {
           if (i > 0) printf (",");
            printf ("%ld", cdfVar->dimSizes[i]);
       }
       printf("], dimVariance = [");
       for (i=0; i < cdfVar->dimensionality; i++) {
           if (i > 0) printf (",");
           if (dimVariance[i] == VARY)
               printf ("VARY");
           else
               printf ("NOVARY,");
       }

       if (cdfVar->recVariance == VARY)
           printf ("], recVary = VARY\n");
       else
           printf ("], recVary = NOVARY\n");
                
   }
   status = CDFlib (CREATE_, zVAR_, cdfVar->name,
                                    cdfVar->datatype,
                                    numElements,
                                    cdfVar->dimensionality,
                                    cdfVar->dimSizes,
                                    cdfVar->recVariance,
                                    dimVariance,
                                    &varNum,
                    NULL_);
   if (DEBUG) printf ("  Returned from CREATE_, zVAR_\n");
   if (status != CDF_OK) cdf_status_handler (status, "CREATE_, zVAR_");

   strcpy (processed_cdf_vars[n_cdf_vars_created].name, cdfVar->name);
   processed_cdf_vars[n_cdf_vars_created++].seq_no = 0;

   return varNum;
}

/*--------------------------------------------------------------------------
 *  This routine returns the string representation of the given CDF
 *  datatype.
 *--------------------------------------------------------------------------*/
char *  cdf_str_datatype (long type)
{
    switch (type) {
      case CDF_BYTE:
         return "CDF_BYTE";

      case CDF_INT1:
         return "CDF_INT1";

      case CDF_CHAR:
         return "CDF_CHAR";

      case CDF_INT2:
         return "CDF_INT2";

      case CDF_UCHAR:
         return "CDF_UCHAR";

      case CDF_UINT1:
         return "CDF_UINT1";

      case CDF_INT4:
         return "CDF_INT4";

      case CDF_INT8:
         return "CDF_INT8";

      case CDF_UINT2:
         return "CDF_UINT2";

      case CDF_FLOAT:
         return "CDF_FLOAT";

      case CDF_REAL4:
         return "CDF_REAL4";

      case CDF_DOUBLE:
         return "CDF_DOUBLE";

      case CDF_REAL8:
         return "CDF_REAL8";

      case CDF_UINT4:
         return "CDF_UINT4";

      case CDF_EPOCH:
         return "CDF_EPOCH";

      case CDF_EPOCH16:
         return "CDF_EPOCH16";

      case CDF_TIME_TT2000:
         return "CDF_TIME_TT2000";

      default:
         return "INVALID CDF data type";
    }
}


/*--------------------------------------------------------------------------*/
/* Write data to the given CDF variable                                     */
/*--------------------------------------------------------------------------*/
void cdf_write_vardata (long varNum,
                        long recNo,      /* Record # from which data is added */
                        long recCount,   /* # of records to be written */
                        long dim,        /* dimensionality     */
                        long dimSizes[], /* dimesion sizes     */
                        void *data)      /* data to be written */
{
   CDFstatus  status;
   int   i, k;
   long  recInterval,                   /* No. of recors to skip            */
         dimIndices[CDF_MAX_DIMS],      /* Beginning location of the array  */
                                        /* to be dumped                     */

         dimIntervals[CDF_MAX_DIMS];    /* No. of elements to skip          */

   recInterval = 1L;             /* # of records to skip between writes   */

   for (i=0; i < dim; i++) {
        dimIndices[i]   = 0;     /* Dump data from the beginning of array */
        dimIntervals[i] = 1;     /* Dump all the elements in the array    */
   }

   status = CDFlib (SELECT_,zVAR_, varNum,
                            zVAR_RECNUMBER_, recNo,
                            zVAR_RECCOUNT_, recCount,
                            zVAR_RECINTERVAL_, recInterval,
                            zVAR_DIMINDICES_, dimIndices,
                            zVAR_DIMCOUNTS_, dimSizes,
                            zVAR_DIMINTERVALS_, dimIntervals,
                    PUT_, zVAR_HYPERDATA_, data,
                    NULL_);

   if (status != CDF_OK)
       cdf_status_handler (status, "PUT_, zVAR_HYPERDATA_, data");
}


/*--------------------------------------------------------------------------*/
/*  Allocate the number of records to initially write to the current        */
/*  zVariable.  This may be specified only once per zVariable and before    */
/*  any data is written.                                                    */
/*--------------------------------------------------------------------------*/
void cdf_allocate_zvar_records (long varNum, long nRecords)
{
   CDFstatus  status;

   status = CDFlib (SELECT_, zVAR_, varNum,
                    PUT_, zVAR_INITIALRECS_, nRecords,
                    NULL_);

   if (status != CDF_OK)
       cdf_status_handler (status, "PUT_, zVAR_INITIALRECS_");
}


/**************************************************************************/
/*  This routine combines the given two strings into one long string and  */
/*  writes that string as a global attribute entry.                       */
/**************************************************************************/
void cdf_add_gAttribute_entry (struct cdfattr *cdfAttr, char *str1, char *str2)
{
   char *str;
   int  str_length;

   str_length = strlen(str1) + strlen(str2);
   str = malloc (sizeof(char) * str_length + 1);
   strcpy (str, str1);
   strcat (str, str2);

   cdfAttr->entryNum++;
   cdf_create_gAttribute_entry (cdfAttr->num,
                                cdfAttr->entryNum,
                                CDF_CHAR,
                                str_length,
                                str);
   free (str);
}

/*************************************************************************/
/*  Create placeholder attributes that are defined in the mapping file.  */
/*************************************************************************/
void cdf_create_placeholder_attributes()
{
    int i;
    long  attr_num;

    for (i=0; i < n_mapping_attrs; i++) {
         if (strcmp(hdf4AttrNames[i], "*") == 0) {
             attr_num = cdf_create_attribute (cdfAttrNames[i], GLOBAL_SCOPE);
             cdf_create_gAttribute_entry (attr_num, 0, CDF_CHAR,
                                          (long) strlen(comments[i]),
                                           comments[i]);
         }
    }
}


/*************************************************************************/
/*  If the mapping file entry has a variable atrribute entry defined,    */
/*  create thedefined attribute (if necessary) and its entry.            */
/*                                                                       */
/*  Example:                                                             */
/*      temp  Temperature  // FORMAT := F11.5                            */
/*                                                                       */
/*      => Will translate 'temp' into "Temperature".  WIll also create   */
/*         an attrubute called FORMAT and assign F11.5 to FORMAT.        */
/*************************************************************************/
void cdf_handle_attrs_in_mapping_file (char *cdfvar_name)
{
    int   i, j, len, create_new_attr, new_attr_name_created;
    long  attr_num;
    char  attr_name[CDF_ATTR_NAME_LEN256+1], attr_value[160];
    char  *ptr, *p1;

    for (i=0; i < n_mapping_attrs; i++) {
         if (strcmp(cdfAttrNames[i], cdfvar_name) == 0) {
             ptr = (char *) strlaststr(comments[i], ":=");
             if (ptr == NULL) return;
             strncpy(attr_name, comments[i], ptr-comments[i]);
             attr_name[ptr-comments[i]] = '\0';
             len = strlen(attr_name);
             if (attr_name[len-1] == ' ') attr_name[len-1] = '\0';

             p1 = comments[i] + strlen(comments[i]);
             strncpy(attr_value, ptr+2, p1-ptr+3);
             attr_value[p1-ptr+2] = '\0';
             len = strlen(attr_value);
             if (attr_value[0] == ' ') strcpy(&attr_value[0], &attr_value[1]);

             cdf_get_new_attribute_name (attr_name, VARIABLE_SCOPE,
                                         &create_new_attr, 
                                         &new_attr_name_created);
   
             if (create_new_attr) 
                 attr_num = cdf_create_attribute (attr_name, VARIABLE_SCOPE);
             
             cdf_create_vAttribute_entry (attr_name,
                                          cdfvar_name,
                                          CDF_CHAR,
                                          strlen(attr_value),
                                          attr_value);
             break;
         }
    }
}

/*************************************************************************/
/*  Set the compression for the variable if its data size is big enough. */
/*************************************************************************/
int cdf_setCompression (struct cdfvar cdfVar, int numRecs) 
{
    CDFstatus status;
    int j, size;
    long dtSize, cParms[]= {6L};

    status = CDFgetDataTypeSize(cdfVar.datatype, &dtSize);
    size = 1;
    size *=  (int) dtSize;
    for (j = 0; j < cdfVar.dimensionality; ++j)
      size *= (int) cdfVar.dimSizes[j];
    size *= numRecs;
    if (size >= 1024) {
      status = CDFlib (SELECT_, zVAR_, cdfVar.num,
                       PUT_, zVAR_COMPRESSION_, GZIP_COMPRESSION,
                                                cParms,
                       NULL_);
      if (status != CDF_OK) cdf_status_handler (status, "PUT_, zVAR_COMPRESSION");
      return 1;
    }
    return 0;
}

/******************************************************************************
* showVersions.
******************************************************************************/

void showVersions(void) {
  long version, release, increment; char subIncrement;
  CDFgetLibraryVersion (&version, &release, &increment, &subIncrement);
  printf("\nhdf4-to-cdf version: 1.0\n");
  printf("  (Based on CDF version %ld.%ld.%ld.%c)\n",version, release,
         increment, subIncrement);
  printf("  (Based on HDF version %s)\n",LIBVER_STRING);
  printf("Released on Dec 14, 2019.\n");
  printf("Submit problems/questions/suggestions to "
         "NASA-CDF-Support@nasa.onmicrosoft.com\n\n");
}

